"""Test module for enabled modules sourced from Github."""
from pathlib import Path

import pytest

from astrality.exceptions import GithubModuleError
from astrality.github import clone_or_pull_repo, clone_repo
from astrality.utils import run_shell


@pytest.mark.slow
def test_clone_github_repo(tmpdir):
    modules_directory = Path(tmpdir.mkdir('modules'))
    repo_dir = clone_repo(
        user='jakobgm',
        repository='color-schemes.astrality',
        modules_directory=modules_directory,
    )
    assert repo_dir.is_dir()

    module_config = repo_dir / 'config.yml'
    assert module_config.is_file()

    assert repo_dir.name == 'color-schemes.astrality'
    assert repo_dir.parent.name == 'jakobgm'


@pytest.mark.slow
def test_cloning_non_existent_github_repository(tmpdir):
    modules_directory = Path(tmpdir.mkdir('modules'))

    with pytest.raises(GithubModuleError):
        clone_repo(
            user='jakobgm',
            repository='i-will-never-create-this-repository',
            modules_directory=modules_directory,
        )

    github_user_directory = modules_directory / 'jakobgm'
    assert not github_user_directory.is_dir()

    repository_directory = github_user_directory / 'i-will-never-create-this-repository'
    assert not repository_directory.is_dir()


@pytest.mark.slow
def test_cloning_two_repositories(tmpdir):
    modules_directory = Path(tmpdir.mkdir('modules'))
    repo_dir = clone_repo(
        user='jakobgm',
        repository='color-schemes.astrality',
        modules_directory=modules_directory,
    )
    repo_dir = clone_repo(
        user='jakobgm',
        repository='solar-desktop.astrality',
        modules_directory=modules_directory,
    )

    github_user_directory = modules_directory / 'jakobgm'
    assert len(tuple(github_user_directory.iterdir())) == 2


@pytest.mark.slow
def test_cloning_one_existent_and_one_non_existent_repo(tmpdir):
    modules_directory = Path(tmpdir.mkdir('modules'))
    repo_dir = clone_repo(
        user='jakobgm',
        repository='color-schemes.astrality',
        modules_directory=modules_directory,
    )

    with pytest.raises(GithubModuleError):
        repo_dir = clone_repo(
            user='jakobgm',
            repository='i-will-never-create-this-repository',
            modules_directory=modules_directory,
        )

    github_user_directory = modules_directory / 'jakobgm'
    assert len(tuple(github_user_directory.iterdir())) == 1


@pytest.mark.slow
def test_cloning_the_same_repo_twice(tmpdir):
    modules_directory = Path(tmpdir.mkdir('modules'))
    repo_dir = clone_repo(
        user='jakobgm',
        repository='color-schemes.astrality',
        modules_directory=modules_directory,
    )

    config_file = modules_directory / 'jakobgm' / 'color-schemes.astrality' / 'config.yml'
    config_file.write_text('user edited')

    repo_dir = clone_repo(
        user='jakobgm',
        repository='color-schemes.astrality',
        modules_directory=modules_directory,
    )

    with open(config_file) as file:
        assert file.read() == 'user edited'

@pytest.mark.slow
def test_clone_or_pull_repository_by_updating_outdated_repository(tmpdir):
    modules_directory = Path(tmpdir.mkdir('modules'))
    repo_dir = clone_repo(
        user='jakobgm',
        repository='color-schemes.astrality',
        modules_directory=modules_directory,
    )

    # Move master to first commit in repository
    result = run_shell(
        command='git reset --hard 2b8941a',
        timeout=5,
        fallback=False,
        working_directory=repo_dir,
    )
    assert result is not False

    # The readme does not exist in this commit
    readme = repo_dir / 'README.rst'
    assert not readme.is_file()

    # The following pull should reintroduce the README into the directory
    updated_repo_dir = clone_or_pull_repo(
        user='jakobgm',
        repository='color-schemes.astrality',
        modules_directory=modules_directory,
    )
    assert updated_repo_dir == repo_dir
    assert readme.is_file()



"""Test module for global module configuration options."""
import shutil
import time
from pathlib import Path

import pytest

from astrality.exceptions import NonExistentEnabledModule
from astrality.config import (
    DirectoryModuleSource,
    EnabledModules,
    GithubModuleSource,
    GlobalModuleSource,
    GlobalModulesConfig,
    ModuleSource,
)
from astrality.utils import run_shell

@pytest.fixture
def modules_application_config():
    return {
        'modules_directory': 'test_modules',
        'enabled_modules': [
            {'name': 'oslo', 'safe': False},
            {'name': 'trondheim'},
        ],
    }


@pytest.yield_fixture(autouse=True)
def delete_jakobgm(test_config_directory):
    """Delete jakobgm module directory used in testing."""
    location1 = test_config_directory / 'freezed_modules' / 'jakobgm'
    location2 = test_config_directory / 'test_modules' / 'jakobgm'

    yield

    if location1.is_dir():
        shutil.rmtree(location1)
    if location2.is_dir():
        shutil.rmtree(location2)


def test_custom_modules_folder(conf_path):
    modules_config = GlobalModulesConfig(
        config={'modules_directory': 'test_modules'},
        config_directory=conf_path,
    )

    assert modules_config.modules_directory == conf_path / 'test_modules'


def test_enabled_modules(test_config_directory):
    modules_config = GlobalModulesConfig({
        'modules_directory': 'test_modules',
        'enabled_modules': [
            {'name': 'oslo::*', 'safe': True},
            {'name': 'trondheim::*'},
        ],
    }, config_directory=test_config_directory)

    # Test that all ExternalModuleSource objects are created
    modules_directory_path = test_config_directory / 'test_modules'
    oslo_path = modules_directory_path / 'oslo'
    trondheim_path = modules_directory_path / 'trondheim'

    oslo = DirectoryModuleSource(
        enabling_statement={'name': 'oslo::*', 'trusted': True},
        modules_directory=modules_directory_path,
    )
    trondheim = DirectoryModuleSource(
        enabling_statement={'name': 'trondheim::*', 'trusted': False},
        modules_directory=modules_directory_path,
    )

    assert len(tuple(modules_config.external_module_sources)) == 2

    assert oslo in tuple(modules_config.external_module_sources)
    assert trondheim in tuple(modules_config.external_module_sources)

    # Test that all module config files are correctly set
    assert oslo_path / 'config.yml' in modules_config.external_module_config_files
    assert trondheim_path / 'config.yml' in modules_config.external_module_config_files

def test_external_module(test_config_directory):
    modules_directory_path = test_config_directory / 'test_modules'
    oslo = DirectoryModuleSource(
        enabling_statement={'name': 'oslo::*'},
        modules_directory=modules_directory_path,
    )

    oslo_path = test_config_directory / 'test_modules' / 'oslo'
    assert oslo.directory == oslo_path
    assert oslo.trusted == True
    assert oslo.relative_directory_path == Path('oslo')
    assert oslo.config_file == oslo_path / 'config.yml'


def test_retrieval_of_external_module_config(test_config_directory):
    external_module_source_config = {'name': 'burma::*'}
    external_module_source = DirectoryModuleSource(
        enabling_statement=external_module_source_config,
        modules_directory=test_config_directory / 'test_modules',
    )

    assert external_module_source.config({}) == {
        f'module/burma::burma': {
            'enabled': True,
            'safe': False,
        },
    }


class TestModuleSource:

    def test_finding_correct_module_source_type_from_name(self):
        assert ModuleSource.type(of='name') == GlobalModuleSource
        assert ModuleSource.type(of='category::name') == DirectoryModuleSource
        assert ModuleSource.type(of='github::user/repo') == GithubModuleSource
        assert ModuleSource.type(of='github::user/repo::module') == GithubModuleSource
        assert ModuleSource.type(of='github::user/repo::*') == GithubModuleSource

    def test_detection_of_all_module_directories_within_a_directory(
        self,
        test_config_directory,
    ):
        assert tuple(EnabledModules.module_directories(
            within=test_config_directory / 'freezed_modules',
        )) == (
            'north_america',
            'south_america',
        )


class TestDirectoryModuleSource:
    """Test of object responsible for module(s) defined in a directory."""

    def test_which_enabling_statements_represents_directory_module_sources(self):
        assert DirectoryModuleSource.represented_by(
            module_name='category::name',
        )
        assert DirectoryModuleSource.represented_by(
            module_name='category/recursive::name',
        )
        assert DirectoryModuleSource.represented_by(
            module_name='category::*',
        )
        assert DirectoryModuleSource.represented_by(
            module_name='*::*',
        )

        assert not DirectoryModuleSource.represented_by(
            module_name='category::*not_valid',
        )
        assert not DirectoryModuleSource.represented_by(
            module_name='category::not_valid*',
        )
        assert not DirectoryModuleSource.represented_by(
            module_name='name',
        )
        assert not DirectoryModuleSource.represented_by(
            module_name='*',
        )

    def test_error_on_enabling_non_existent_module(
        self,
        test_config_directory,
    ):
        enabling_statement = {'name': 'north_america::colombia'}
        with pytest.raises(NonExistentEnabledModule):
            directory_module = DirectoryModuleSource(
                enabling_statement=enabling_statement,
                modules_directory=test_config_directory / 'freezed_modules',
            )
            config = directory_module.config({})

    def test_recursive_module_directory(
        self,
        test_config_directory,
    ):
        enabling_statement = {'name': 'recursive/directory::bulgaria'}
        directory_module = DirectoryModuleSource(
            enabling_statement=enabling_statement,
            modules_directory=test_config_directory / 'test_modules',
        )
        assert directory_module.config({}) == {
            'module/recursive/directory::bulgaria': {
                'on_startup': {
                    'run': "echo 'Greetings from Bulgaria!'",
                },
            },
        }

    def test_getting_config_dict_from_directory_module(
        self,
        test_config_directory,
    ):
        enabling_statement = {'name': 'north_america::USA'}
        directory_module = DirectoryModuleSource(
            enabling_statement=enabling_statement,
            modules_directory=test_config_directory / 'freezed_modules',
        )
        assert directory_module.config({}) == {
            'module/north_america::USA': {
                'on_startup': {
                    'run': 'echo Greetings from the USA!',
                },
            },
            'context/geography': {
                'USA': {
                    'capitol': 'Washington D.C.',
                },
            },
        }

    def test_enabling_just_one_module_in_diretory(
        self,
        test_config_directory,
    ):
        enabling_statement = {'name': 'south_america::brazil'}
        directory_module = DirectoryModuleSource(
            enabling_statement=enabling_statement,
            modules_directory=test_config_directory / 'freezed_modules',
        )
        assert directory_module.config({}) == {
            'module/south_america::brazil': {
                'on_startup': {
                    'run': 'echo Greetings from Brazil!',
                },
            },
            'context/geography': {
                'brazil': {
                    'capitol': 'Brasilia',
                },
                'argentina': {
                    'capitol': 'Buenos Aires',
                },
            },
        }

    def test_checking_if_directory_modules_contain_module_name(
        self,
        test_config_directory,
    ):
        enabling_statement = {'name': 'south_america::brazil'}
        directory_module = DirectoryModuleSource(
            enabling_statement=enabling_statement,
            modules_directory=test_config_directory / 'freezed_modules',
        )
        directory_module.config({})

        assert 'south_america::brazil' in directory_module

        assert 'south_america::argentina' not in directory_module
        assert 'brazil' not in directory_module
        assert 'argentina' not in directory_module


class TestGlobalModuleSource:

    def test_valid_names_which_indicate_globally_defined_modules(self):
        assert GlobalModuleSource.represented_by(
            module_name='module_name',
        )
        assert GlobalModuleSource.represented_by(
            module_name='*',
        )

        assert not GlobalModuleSource.represented_by(
            module_name='w*',
        )
        assert not GlobalModuleSource.represented_by(
            module_name='category::name',
        )
        assert not GlobalModuleSource.represented_by(
            module_name='category::*',
        )
        assert not GlobalModuleSource.represented_by(
            module_name='*::*',
        )
        assert not GlobalModuleSource.represented_by(
            module_name='jakobgm/module',
        )

    def test_that_enabled_modules_are_detected_correctly(self):
        global_module_source = GlobalModuleSource(
            enabling_statement={'name': 'enabled_module'},
            modules_directory=Path('/'),
        )
        global_module_source.config({})
        assert 'enabled_module' in global_module_source
        assert 'disabled_module' not in global_module_source
        assert '*' not in global_module_source

    def test_that_wildcard_global_enabling_enables_all_global_modules(self):
        global_module_source = GlobalModuleSource(
            enabling_statement={'name': '*'},
            modules_directory=Path('/'),
        )
        assert 'enabled_module1' in global_module_source
        assert 'enabled_module2' in global_module_source

        assert 'directory::module' not in global_module_source
        assert 'user/repo' not in global_module_source


class TestGithubModuleSource:

    def test_valid_names_which_indicate_github_modules(self):
        assert GithubModuleSource.represented_by(
            module_name='github::jakobgm/astrality',
        )
        assert GithubModuleSource.represented_by(
            module_name='github::jakobgm/astrality::module',
        )
        assert GithubModuleSource.represented_by(
            module_name='github::jakobgm/astrality::*',
        )
        assert GithubModuleSource.represented_by(
            module_name='github::user_name./repo-git.ast',
        )

        assert not GithubModuleSource.represented_by(
            module_name='w*',
        )
        assert not GithubModuleSource.represented_by(
            module_name='category::name',
        )
        assert not GithubModuleSource.represented_by(
            module_name='category::*',
        )
        assert not GithubModuleSource.represented_by(
            module_name='*::*',
        )
        assert not GithubModuleSource.represented_by(
            module_name='*',
        )
        assert not GithubModuleSource.represented_by(
            module_name='global_module',
        )

    def test_that_username_and_repo_is_identified(self, tmpdir, delete_jakobgm):
        modules_directory = Path(tmpdir)
        github_module_source = GithubModuleSource(
            enabling_statement={'name': 'github::jakobgm/astrality'},
            modules_directory=modules_directory,
        )
        assert github_module_source.github_user == 'jakobgm'
        assert github_module_source.github_repo == 'astrality'

    @pytest.mark.slow
    def test_that_enabled_repos_are_found(self, test_config_directory, delete_jakobgm):
        github_module_source = GithubModuleSource(
            enabling_statement={'name': 'github::jakobgm/test-module.astrality'},
            modules_directory=test_config_directory / 'test_modules',
        )
        github_module_source.config({})

        assert 'github::jakobgm/test-module.astrality::botswana' in github_module_source
        assert 'github::jakobgm/test-module.astrality::ghana' in github_module_source

        assert 'github::jakobgm/test-module.astrality::non_existent' not in github_module_source
        assert 'github::jakobgm/another_repo::ghana' not in github_module_source
        assert 'astrality' not in github_module_source
        assert 'jakobgm' not in github_module_source

    @pytest.mark.slow
    def test_specific_github_modules_enabled(self, test_config_directory, delete_jakobgm):
        github_module_source = GithubModuleSource(
            enabling_statement={'name': 'github::jakobgm/test-module.astrality::botswana'},
            modules_directory=test_config_directory / 'test_modules',
        )
        github_module_source.config({})

        assert 'github::jakobgm/test-module.astrality::botswana' in github_module_source
        assert 'github::jakobgm/test-module.astrality::ghana' not in github_module_source

    @pytest.mark.slow
    def test_that_all_modules_enabled_syntaxes_behave_identically(
        self,
        test_config_directory,
        delete_jakobgm,
    ):
        github_module_source1 = GithubModuleSource(
            enabling_statement={'name': 'github::jakobgm/test-module.astrality'},
            modules_directory=test_config_directory / 'test_modules',
        )
        github_module_source1.config({})

        # Sleep to prevent race conditions
        time.sleep(1)

        github_module_source2 = GithubModuleSource(
            enabling_statement={'name': 'github::jakobgm/test-module.astrality::*'},
            modules_directory=test_config_directory / 'test_modules',
        )
        github_module_source2.config({})

        assert github_module_source1 == github_module_source2

    @pytest.mark.slow
    def test_automatical_retrival_of_github_module(self, tmpdir):
        modules_directory = Path(tmpdir)
        github_module_source = GithubModuleSource(
            enabling_statement={
                'name': 'github::jakobgm/test-module.astrality::*',
                'autoupdate': True,
            },
            modules_directory=modules_directory,
        )
        assert github_module_source.config({}) == {
            'module/github::jakobgm/test-module.astrality::botswana': {
                'on_startup': {
                    'run': "echo 'Greetings from Botswana!'",
                },
            },
            'module/github::jakobgm/test-module.astrality::ghana': {
                'on_startup': {
                    'run': "echo 'Greetings from Ghana!'",
                },
            },
            'context/geography': {
                'botswana': {
                    'capitol': 'Gaborone',
                },
                'ghana': {
                    'capitol': 'Accra',
                },
            },
        }

    @pytest.mark.slow
    def test_use_of_autoupdating_github_source(self, tmpdir):
        modules_directory = Path(tmpdir)

        github_module_source = GithubModuleSource(
            enabling_statement={
                'name': 'github::jakobgm/test-module.astrality',
                'autoupdate': True,
            },
            modules_directory=modules_directory,
        )

        # The repository is lazely cloned, so we need to get the config
        config = github_module_source.config({})

        repo_dir = modules_directory / 'jakobgm' / 'test-module.astrality'
        assert repo_dir.is_dir()

        # Move master to first commit in repository
        result = run_shell(
            command='git reset --hard d4c9723',
            timeout=5,
            fallback=False,
            working_directory=repo_dir,
        )
        assert result is not False

        # The readme does not exist in this commit
        readme = repo_dir / 'README.rst'
        assert not readme.is_file()

        del github_module_source
        github_module_source = GithubModuleSource(
            enabling_statement={
                'name': 'github::jakobgm/test-module.astrality',
                'autoupdate': True,
            },
            modules_directory=modules_directory,
        )
        config = github_module_source.config({})

        # The autoupdating should update the module to origin/master
        # containing the README.rst file
        assert readme.is_file()


class TestEnabledModules:

    def test_processing_of_enabled_statements(
        self,
        test_config_directory,
    ):
        enabled_modules = EnabledModules([], Path('/'), Path('/'))

        assert enabled_modules.process_enabling_statements(
            enabling_statements=[
                {'name': '*::*'}
            ],
            modules_directory=test_config_directory / 'freezed_modules',
        ) == [{'name': 'north_america::*'}, {'name': 'south_america::*'}]
        assert enabled_modules.all_directory_modules_enabled == True
        assert enabled_modules.all_global_modules_enabled == False


    @pytest.mark.slow
    def test_enabled_detection(
        self,
        test_config_directory,
        caplog,
        delete_jakobgm,
    ):
        enabling_statements = [
            {'name': 'global'},
            {'name': 'south_america::*'},
            {'name': 'github::jakobgm/test-module.astrality'},
            {'name': 'invalid_syntax]][['},
        ]
        enabled_modules = EnabledModules(
            enabling_statements=enabling_statements,
            config_directory=test_config_directory,
            modules_directory=test_config_directory / 'freezed_modules',
        )

        for sources in enabled_modules.source_types.values():
            assert len(sources) == 1
        assert len(caplog.record_tuples) == 1

        enabled_modules.compile_config_files({})
        assert 'global' in enabled_modules
        assert 'south_america::brazil' in enabled_modules
        assert 'south_america::argentina' in enabled_modules
        assert 'github::jakobgm/test-module.astrality::botswana' in enabled_modules
        assert 'github::jakobgm/test-module.astrality::ghana' in enabled_modules

        assert 'not_enabled' not in enabled_modules
        assert 'non_existing_folder::non_existing_module' not in enabled_modules
        assert 'github::user/not_enabled' not in enabled_modules

    def test_enabled_detection_with_global_wildcard(self):
        enabling_statements = [
            {'name': '*'},
        ]
        enabled_modules = EnabledModules(
            enabling_statements=enabling_statements,
            config_directory=Path('/'),
            modules_directory=Path('/'),
        )

        assert 'global' in enabled_modules
        assert 'whatever' in enabled_modules

        assert 'south_america::brazil' not in enabled_modules
        assert 'south_america::argentina' not in enabled_modules
        assert 'github::jakobgm/color_schemes.astrality' not in enabled_modules



"""Module implementing user configured custom functionality."""

from collections import namedtuple
from datetime import timedelta
import logging
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from jinja2.exceptions import TemplateNotFound

from astrality import compiler
from astrality.compiler import context
from astrality.config import (
    ApplicationConfig,
    GlobalModulesConfig,
    expand_path,
    insert_into,
    user_configuration,
)
from astrality.event_listener import EventListener, event_listener_factory
from astrality.filewatcher import DirectoryWatcher
from astrality.resolver import Resolver
from astrality.utils import run_shell


ModuleConfig = Dict[str, Any]

ContextSectionImport = namedtuple(
    'ContextSectionImport',
    ['into_section', 'from_section', 'from_config_file'],
)
Template = namedtuple(
    'Template',
    ['source', 'target'],
)
WatchedFile = namedtuple(
    'WatchedFile',
    ['path', 'module', 'specified_path'],
)

logger = logging.getLogger('astrality')


class Module:
    """
    Class for determining module actions.

    A Module instance should only return actions to be performed, but never
    perform them. That is the responsibility of a ModuleManager instance.

    A module can define a set of commands to be run on astrality startup, and
    exit, in addition to every time a given type of event changes.

    Commands are run in the users shell, and can use the following placeholder:
    - {event}: The event specified by the event_listener instance.
    """

    def __init__(
        self,
        module_config: ModuleConfig,
        module_directory: Path,
    ) -> None:
        """
        Initialize Module object with a section from a config dictionary.

        Section name must be [module/*], where * is the module name.
        In addition, the enabled option must be set to "true", or not set at
        all.

        module_config example:
        {'module/name':
            'enabled': True,
            'event_listener': {'type': 'weekday'},
            'on_startup': {'run': ['echo weekday is {event}']},
        }
        """
        # Can only initialize one module at a time
        assert len(module_config) == 1

        section = next(iter(module_config.keys()))
        self.name: str = section[7:]

        # The source directory for the module, determining how to interpret
        # relative paths in the module config
        self.directory = module_directory

        self.module_config = module_config[section]
        self.populate_event_blocks()

        # Import trigger actions into their respective event blocks
        self.import_trigger_actions()

        # Use static event_listener if no event_listener is specified
        self.event_listener: EventListener = event_listener_factory(
            self.module_config.get('event_listener', {'type': 'static'}),
        )

    def populate_event_blocks(self) -> None:
        """
        Populate non-configured actions within event blocks.

        This prevents us from having to use .get() all over the Module.
        """
        for event_block in ('on_startup', 'on_event', 'on_exit', ):
            configured_event_block = self.module_config.get(event_block, {})
            self.module_config[event_block] = {
                'import_context': [],
                'compile': [],
                'run': [],
                'trigger': [],
            }
            self.module_config[event_block].update(configured_event_block)

        if not 'on_modified' in self.module_config:
            self.module_config['on_modified'] = {}
        else:
            for template_name in self.module_config['on_modified'].keys():
                configured_event_block = self.module_config['on_modified'][template_name]
                self.module_config['on_modified'][template_name] = {
                    'import_context': [],
                    'compile': [],
                    'run': [],
                    'trigger': [],
                }
                self.module_config['on_modified'][template_name].update(configured_event_block)

        # Convert any single actions into a list of that action, allowing
        # users to not use lists in their configuration if they only have one
        # action
        for event_block in (
            self.module_config['on_startup'],
            self.module_config['on_event'],
            self.module_config['on_exit'],
            *self.module_config['on_modified'].values(),
        ):
            for action in ('import_context', 'compile', 'run', 'trigger',):
                if not isinstance(event_block[action], list):  # type: ignore
                    event_block[action] = [event_block[action]]  # type: ignore

    def import_trigger_actions(self) -> None:
        """If an event block defines trigger events, import those actions."""
        event_blocks = (
            self.module_config['on_startup'],
            self.module_config['on_event'],
            self.module_config['on_exit'],
            *self.module_config['on_modified'].values(),
        )
        for event_block in event_blocks:
            event_blocks_to_import = event_block['trigger']

            for event_block_to_import in event_blocks_to_import:
                self._import_event_block(
                    from_event_block=event_block_to_import,
                    into=event_block,
                )

    def _import_event_block(
        self,
        from_event_block: str,
        into: Dict[str, Any],
    ) -> None:
        """Merge one event block with another one."""
        if 'on_modified:' in from_event_block:
            template = from_event_block[12:]
            from_event_block_dict = self.module_config['on_modified'].get(template, {})
        else:
            from_event_block_dict = self.module_config[from_event_block]

        into['run'].extend(from_event_block_dict['run'])
        into['import_context'].extend(from_event_block_dict['import_context'])
        into['compile'].extend(from_event_block_dict['compile'])

    def commands(
        self,
        block: str,
        modified_file: Optional[str] = None,
    ) -> Tuple[str, ...]:
        """
        Return all shell commands to be run for a specific block, i.e.
        any of on_startup, on_event, on_exit, or on_modified.

        A modified file action block is given by block='on_modified:file/path'.
        """

        startup_commands: List[str]

        if modified_file:
            assert block == 'on_modified'
            startup_commands = self.module_config['on_modified'][modified_file]['run']
        else:
            assert block in ('on_startup', 'on_event', 'on_exit',)
            startup_commands = self.module_config[block]['run']

        if len(startup_commands) == 0:
            logger.debug(f'[module/{self.name}] No {block} command specified.')
            return ()
        else:
            return tuple(
                self.interpolate_string(command)
                for command
                in startup_commands
            )

    def startup_commands(self) -> Tuple[str, ...]:
        """Return commands to be run on Module instance startup."""
        return self.commands('on_startup')

    def on_event_commands(self) -> Tuple[str, ...]:
        """Commands to be run when self.event_listener event changes."""
        return self.commands('on_event')

    def exit_commands(self) -> Tuple[str, ...]:
        """Commands to be run on Module instance shutdown."""
        return self.commands('on_exit')

    def modified_commands(self, specified_path: str) -> Tuple[str, ...]:
        """Commands to be run when a module template is modified."""
        return self.commands('on_modified', specified_path)

    def context_section_imports(
        self,
        trigger: str,
        modified: Optional[str] = None,
    ) -> Tuple[ContextSectionImport, ...]:
        """
        Return what to import into the global application_context.

        Trigger is one of 'on_startup', 'on_event', 'on_modified, or 'on_exit'.
        This determines which section of the module is used to get the context
        import specification from.

        If trigger is 'on_modified', you also need to specify which file is
        modified, in order to get the correct section.
        """
        import_config: List[Dict[str, str]]
        if modified:
            assert trigger == 'on_modified'
            import_config = self.module_config[trigger][modified]['import_context']
        else:
            assert trigger in ('on_startup', 'on_event', 'on_exit',)
            import_config = self.module_config[trigger]['import_context']

        context_section_imports = []
        for context_import in import_config:
            # Insert placeholders
            from_path = self.interpolate_string(context_import['from_path'])

            from_section: Optional[str]
            to_section: Optional[str]

            if 'from_section' in context_import:
                from_section = self.interpolate_string(context_import['from_section'])

                # If no `to_section` is specified, use the same section as
                # `from_section`
                to_section = self.interpolate_string(
                    context_import.get('to_section', from_section),
                )
            else:
                # From section is not specified, so set both to None, indicating
                # a wish to import *all* sections
                from_section = None
                to_section = None

            # Get the relative path to file containing the context
            config_path = Path(from_path)

            # Isert a ContextSectionImport tuple into the return value
            context_section_imports.append(
                ContextSectionImport(
                    into_section=to_section,
                    from_section=from_section,
                    from_config_file=config_path,
                )
            )

        return tuple(context_section_imports)


    def interpolate_string(self, string: str) -> str:
        """
        Replace all module placeholders in string.

        For now, the module only replaces {event} placeholders.
        """
        return string.replace('{event}', self.event_listener.event())

    @staticmethod
    def valid_class_section(
        section: ModuleConfig,
        requires_timeout: Union[int, float],
        requires_working_directory: Path,
    ) -> bool:
        """Check if the given dict represents a valid enabled module."""

        if not len(section) == 1:
            raise RuntimeError(
                'Tried to check module section with dict '
                'which does not have exactly one item.',
            )

        try:
            module_name = next(iter(section.keys()))
            valid_module_name = module_name.split('/')[0].lower() == 'module'  # type: ignore
            enabled = section[module_name].get('enabled', True)
            if not (valid_module_name and enabled):
                return False

        except KeyError:
            return False

        # The module is enabled, now check if all requirements are satisfied
        requires = section[module_name].get('requires')
        if not requires:
            return True
        else:
            if isinstance(requires, str):
                requires = [requires]

            for requirement in requires:
                if run_shell(command=requirement, fallback=False) is False:
                    logger.warning(
                        f'[{module_name}] Module does not satisfy requirement "{requirement}".',
                    )
                    return False

            logger.info(
                f'[{module_name}] Module satisfies all requirements.'
            )
            return True

class ModuleManager:
    """A manager for operating on a set of modules."""

    def __init__(self, config: ApplicationConfig) -> None:
        """Initialize a ModuleManager object from `astrality.yml` dict."""

        self.config_directory = Path(config['_runtime']['config_directory'])
        self.temp_directory = Path(config['_runtime']['temp_directory'])
        self.application_config = config
        self.application_context: Dict[str, Resolver] = {}

        self.startup_done = False
        self.last_module_events: Dict[str, str] = {}

        # Get module configurations which are externally defined
        self.global_modules_config = GlobalModulesConfig(  # type: ignore
            config=config.get('config/modules', {}),
            config_directory=self.config_directory,
        )
        self.recompile_modified_templates = self.global_modules_config.recompile_modified_templates

        self.modules: Dict[str, Module] = {}

        # Application context is used in compiling external config sources
        application_context = context(config)

        # Insert externally managed modules
        for external_module_source in self.global_modules_config.external_module_sources:
            module_directory = external_module_source.directory

            module_configs = external_module_source.config(
                context=application_context,
            )

            # Insert context defined in external configuration
            self.application_context.update(context(module_configs))

            for section, options in module_configs.items():
                module_config = {section: options}

                if not Module.valid_class_section(
                    section=module_config,
                    requires_timeout=self.global_modules_config.requires_timeout,
                    requires_working_directory=module_directory,
                ) or section not in self.global_modules_config.enabled_modules:
                    continue

                module = Module(
                    module_config=module_config,
                    module_directory=module_directory,
                )
                self.modules[module.name] = module

        # Update the context from `astrality.yml`, overwriting any defined
        # contexts in external modules in the case of naming conflicts
        self.application_context.update(application_context)

        # Insert modules defined in `astrality.yml`
        for section, options in config.items():
            module_config = {section: options}

            # Check if this module should be included
            if not Module.valid_class_section(
                section=module_config,
                requires_timeout=self.global_modules_config.requires_timeout,
                requires_working_directory=self.config_directory,
            ) or section not in self.global_modules_config.enabled_modules:
                continue

            module = Module(
                module_config=module_config,
                module_directory=self.config_directory,
            )
            self.modules[module.name] = module

        self.templates = self.prepare_templates(self.modules.values())
        self.string_replacements = self.generate_string_replacements(self.templates)
        self.on_modified_paths = self.find_on_modified_paths(self.modules.values())

        # Initialize the config directory watcher, but don't start it yet
        self.directory_watcher = DirectoryWatcher(
            directory=self.config_directory,
            on_modified=self.file_system_modified,
        )

        logger.info('Enabled modules: ' + ', '.join(self.modules.keys()))

    def __len__(self) -> int:
        """Return the number of managed modules."""
        return len(self.modules)

    def prepare_templates(
        self,
        modules: Iterable[Module],
    ) -> Dict[str, Template]:
        """
        Prepare the use of templates that could be compiled by `modules`.

        Returns a dictionary where the user provided path to the template is
        the key, and the value is a Template NamedTuple instance, containing
        the source and target Paths of the template.
        """
        templates: Dict[str, Template] = {}

        for module in modules:
            # All the module blocks which can contain compile actions
            for block in (
                    module.module_config['on_startup'],
                    module.module_config['on_event'],
                    module.module_config['on_exit'],
                    *module.module_config['on_modified'].values(),
            ):
                for compile_action in block['compile']:
                    specified_source = compile_action['template']
                    absolute_source = expand_path(
                        path=Path(specified_source),
                        config_directory=module.directory,
                    )

                    if 'target' in compile_action:
                        target = expand_path(
                            path=Path(compile_action['target']),
                            config_directory=module.directory,
                        )
                    else:
                        target = self.create_temp_file(name=module.name)

                    templates[specified_source] = Template(
                        source=absolute_source,
                        target=target,
                    )

        return templates


    def find_on_modified_paths(
        self,
        modules: Iterable[Module],
    ) -> Dict[Path, WatchedFile]:
        """
        Return a dictionary keyed to all modification watched file paths.

        Return a dictionary with the paths to watched files as the keys and
        WatchedFile instances as values.
        """
        on_modified_paths: Dict[Path, WatchedFile] = {}

        for module in modules:
            for watched_for_modification in module.module_config['on_modified'].keys():
                on_modified_path = expand_path(
                    path=Path(watched_for_modification),
                    config_directory=module.directory,
                )
                on_modified_paths[on_modified_path] = WatchedFile(
                    path=on_modified_path,
                    module=module,
                    specified_path=watched_for_modification,
                )

        return on_modified_paths

    def generate_string_replacements(
        self,
        templates: Dict[str, Template],
    ) -> Dict[str, str]:
        """
        Returns a dictionary containing all string replacements keyed to their
        placeholders.

        Includes template path placeholders replaced with the compilation
        target path.
        """
        string_replacements: Dict[str, str] = {}

        for specified_path, template in templates.items():
            string_replacements[specified_path] = str(template.target)

        return string_replacements

    def module_events(self) -> Dict[str, str]:
        """Return dict containing the event of all modules."""
        module_events = {}
        for module_name, module in self.modules.items():
            module_events[module_name] = module.event_listener.event()

        return module_events

    def finish_tasks(self) -> None:
        """
        Finish all due tasks defined by the managed modules.

        The order of finishing tasks is as follows:
            1) Import any relevant context sections.
            2) Compile all templates with the new section.
            3) Run startup commands, if it is not already done.
            4) Run on_event commands, if it is not already done for this
               module events combination.
        """
        if not self.startup_done:
            # Save the last event configuration, such that on_event
            # is only run when the event *changes*
            self.last_module_events = self.module_events()

            # Perform all startup actions
            self.import_context_sections('on_startup')
            self.compile_templates('on_startup')
            self.startup()
        elif self.last_module_events != self.module_events():
            # One or more module events have changed, execute the event blocks
            # of these modules.

            for module_name, event in self.module_events().items():
                if not self.last_module_events[module_name] == event:
                    # This module has a new event

                    self.import_context_sections(
                        trigger='on_event',
                        module=self.modules[module_name],
                    )
                    self.compile_templates(
                        trigger='on_event',
                        module=self.modules[module_name],
                    )
                    self.run_on_event_commands(
                        module=self.modules[module_name],
                    )

                    # Save the event
                    self.last_module_events[module_name] = event

    def has_unfinished_tasks(self) -> bool:
        """Return True if there are any module tasks due."""
        if not self.startup_done:
            return True
        else:
            return self.last_module_events != self.module_events()

    def time_until_next_event(self) -> timedelta:
        """Time left until first event change of any of the modules managed."""
        return min(
            module.event_listener.time_until_next_event()
            for module
            in self.modules.values()
        )

    def import_context_sections(
        self,
        trigger: str,
        module: Optional[Module] = None,
    ) -> None:
        """
        Import context sections defined by the managed modules.

        Trigger is one of 'on_startup', 'on_event', or 'on_exit'.
        This determines which event block of the module is used to get the
        context import specification from.
        """
        assert trigger in ('on_startup', 'on_event', 'on_exit',)

        modules: Iterable[Module]
        if isinstance(module, Module):
            modules = (module, )
        else:
            modules = self.modules.values()

        for module in modules:
            context_section_imports = module.context_section_imports(trigger)

            for csi in context_section_imports:
                self.application_context = insert_into(
                    context=self.application_context,
                    section=csi.into_section,
                    from_section=csi.from_section,
                    from_config_file=expand_path(
                        path=csi.from_config_file,
                        config_directory=module.directory,
                    ),
                )

    def compile_templates(
        self,
        trigger: str,
        module: Optional[Module] = None,
    ) -> None:
        """
        Compile the module templates specified by the `templates` option.

        Trigger is one of 'on_startup', 'on_event', or 'on_exit'.
        This determines which section of the module is used to get the compile
        specification from.
        """
        assert trigger in ('on_startup', 'on_event', 'on_exit',)

        modules: Iterable[Module]
        if isinstance(module, Module):
            modules = (module, )
        else:
            modules = self.modules.values()


        for module in self.modules.values():
            for compilation in module.module_config[trigger]['compile']:
                specified_path = compilation['template']
                template = self.templates[specified_path]
                self.compile_template(
                    source=template.source,
                    target=template.target,
                )


    def compile_template(self, source: Path, target: Path) -> None:
        """
        Compile a single template given by its shortname.

        A shortname is given either by shortname, implying that the module given
        defines that template, or by module_name.shortname, making it explicit.
        """
        try:
            compiler.compile_template(
                template=source,
                target=target,
                context=self.application_context,
                shell_command_working_directory=self.config_directory,
            )
        except TemplateNotFound:
            logger.error(
                f'Could not compile template "{source}" to target "{target}". '
                'Template does not exist.'
            )

    def startup(self):
        """Run all startup commands specified by the managed modules."""
        for module in self.modules.values():
            startup_commands = module.startup_commands()
            for command in startup_commands:
                logger.info(f'[module/{module.name}] Running startup command.')
                self.run_shell(
                    command=command,
                    timeout=self.global_modules_config.run_timeout,
                    working_directory=module.directory,
                    module_name=module.name,
                )

        self.startup_done = True

        # Start watching config directory for file changes
        self.directory_watcher.start()

    def run_on_event_commands(
        self,
        module: Module,
    ):
        """Run all event change commands specified by a managed module."""
        on_event_commands = module.on_event_commands()
        for command in on_event_commands:
            logger.info(f'[module/{module.name}] Running event command.')
            self.run_shell(
                command=command,
                timeout=self.global_modules_config.run_timeout,
                working_directory=module.directory,
                module_name=module.name,
            )

    def exit(self):
        """
        Run all exit tasks specified by the managed modules.

        Also close all temporary file handlers created by the modules.
        """
        # First import context and compile templates
        self.import_context_sections('on_exit')
        self.compile_templates('on_exit')

        # Then run all shell commands
        for module in self.modules.values():
            exit_commands = module.exit_commands()
            for command in exit_commands:
                logger.info(f'[module/{module.name}] Running exit command.')
                self.run_shell(
                    command=command,
                    timeout=self.global_modules_config.run_timeout,
                    working_directory=module.directory,
                    module_name=module.name,
                )

        if hasattr(self, 'temp_files'):
            for temp_file in self.temp_files:
                temp_file.close()

            # Prevent files from being closed again
            del self.temp_files

        # Stop watching config directory for file changes
        self.directory_watcher.stop()

    def on_modified(self, modified: Path) -> None:
        """
        Perform actions when a watched file is modified.
        """
        watched_file = self.on_modified_paths[modified]
        module = watched_file.module
        specified_path = watched_file.specified_path

        # First import context sections in on_modified block
        for csi in module.context_section_imports(
            trigger='on_modified',
            modified=specified_path,
        ):
            self.application_context = insert_into(
                context=self.application_context,
                section=csi.into_section,
                from_section=csi.from_section,
                from_config_file=expand_path(
                    path=csi.from_config_file,
                    config_directory=module.directory,
                ),
            )

        # Now compile templates specified in on_modified block
        for compilation in module.module_config['on_modified'][specified_path]['compile']:
            template = self.templates[compilation['template']]
            self.compile_template(
                source=template.source,
                target=template.target,
            )


        # Lastly, run commands specified in on_modified block
        modified_commands = module.modified_commands(specified_path)
        for command in modified_commands:
            logger.info(f'[module/{module.name}] Running modified command.')
            self.run_shell(
                command=command,
                timeout=self.global_modules_config.run_timeout,
                working_directory=module.directory,
                module_name=module.name,
            )


    def file_system_modified(self, modified: Path) -> None:
        """
        Callback for when files within the config directory are modified.

        Run any context imports, compilations, and shell commands specified
        within the on_modified event block of each module.

        Also, if hot_reload is True, we reinstantiate the ModuleManager object
        if the application configuration has been modified.
        """
        if modified == self.application_config['_runtime']['config_directory'] / 'astrality.yml':
            self.on_application_config_modified()
        elif modified in self.on_modified_paths:
            # The modified file is specified in one of the modules
            self.on_modified(modified)
        else:
            # Check if the modified path is a template which is supposed to
            # be recompiled.
            self.recompile_modified_template(modified=modified)


    def on_application_config_modified(self):
        """
        Reload the ModuleManager if astrality.yml has been modified.

        Reloadnig the module manager only occurs if the user has configured
        `hot_reload_config`.
        """
        if not self.application_config['config/astrality']['hot_reload_config']:
            # Hot reloading is not enabled, so we return early
            return

        # Hot reloading is enabled, get the new configuration dict
        new_application_config = user_configuration(
            config_directory=self.config_directory,
        )

        try:
            # Reinstantiate this object
            new_module_manager = ModuleManager(new_application_config)

            # Run all old exit actions, since the new config is valid
            self.exit()

            # Swap place with the new configuration
            self = new_module_manager

            # Run startup commands from the new configuration
            self.finish_tasks()
        except:
            # New configuration is invalid, just keep the old one
            # TODO: Test this behaviour
            logger.error('New configuration detected, but it is invalid!')
            pass


    def recompile_modified_template(self, modified: Path):
        """
        Recompile any modified template if configured.

        This requires setting the global setting:
        recompile_modified_templates: true
        """
        if not self.recompile_modified_templates:
            return

        for template in self.templates.values():
            if template.source == modified:
                self.compile_template(
                    source=template.source,
                    target=template.target,
                )

    def run_shell(
        self,
        command: str,
        timeout: Union[int, float],
        working_directory: Path,
        module_name: Optional[str] = None,
    ) -> None:
        """Run a shell command defined by a managed module."""

        command = self.interpolate_string(command)

        if module_name:
            logger.info(f'[module/{module_name}] Running command "{command}".')

        run_shell(
            command=command,
            timeout=timeout,
            working_directory=working_directory,
        )

    def interpolate_string(self, string: str) -> str:
        """Replace all template placeholders with the compilation path."""

        for specified_path, template in self.templates.items():
            string = string.replace(
                '{' + specified_path + '}',
                str(template.target),
            )
        return string

    def create_temp_file(self, name) -> Path:
        """Create a temp file used as a compilation target, returning its path."""

        temp_file = NamedTemporaryFile(  # type: ignore
            prefix=name + '-',
            # dir=Path(self.temp_directory),
        )

        # NB: These temporary files need to be persisted during the entirity of
        # the scripts runtime, since the files are deleted when they go out of
        # scope.
        if not hasattr(self, 'temp_files'):
            self.temp_files = [temp_file]
        else:
            self.temp_files.append(temp_file)

        return Path(temp_file.name)



"""Application wide fixtures."""
import copy
import os
import shutil
from pathlib import Path

import pytest

from astrality.config import (
    ASTRALITY_DEFAULT_GLOBAL_SETTINGS,
    user_configuration,
)
from astrality.utils import generate_expanded_env_dict


@pytest.fixture
def conf_path():
    """Return str path to configuration directory."""
    conf_path = Path(__file__).parents[1] / 'config'
    return conf_path


@pytest.fixture
def conf_file_path(conf_path):
    """Return path to example configuration."""
    return conf_path / 'astrality.yml'


@pytest.fixture(scope='session', autouse=True)
def conf():
    """Return the configuration object for the example configuration."""
    this_test_file = os.path.abspath(__file__)
    conf_path = Path(this_test_file).parents[1] / 'config'
    return user_configuration(conf_path)


@pytest.fixture
def expanded_env_dict():
    """Return expanded environment dictionary."""
    return generate_expanded_env_dict()


@pytest.fixture
def default_global_options():
    """Return dictionary containing all default global options."""
    return copy.deepcopy(ASTRALITY_DEFAULT_GLOBAL_SETTINGS)


@pytest.fixture
def _runtime(temp_directory, test_config_directory):
    return {'_runtime': {
        'config_directory': test_config_directory,
        'temp_directory': temp_directory,
    }}


@pytest.fixture
def test_config_directory():
    """Return path to test config directory."""
    return Path(__file__).parent / 'test_config'


@pytest.yield_fixture
def temp_directory():
    """Return path to temporary directory, and cleanup afterwards."""
    temp_dir = Path('/tmp/astrality')
    if not temp_dir.is_dir():
        os.makedirs(temp_dir)

    yield temp_dir

    # Cleanup temp dir after test has been run
    shutil.rmtree(temp_dir, ignore_errors=True)



"""Tests for everything related to module compile actions."""
import os
from pathlib import Path

import pytest

from astrality.module import ModuleManager, WatchedFile

@pytest.yield_fixture
def compiling_fixtures(
    default_global_options,
    _runtime,
    test_config_directory,
):
    """Return a set of fixtures required to inspect template compilation."""
    fixtures = {}

    fixtures['template1'] = Path('template1')
    fixtures['template2'] = Path('template2')
    fixtures['template2_target'] = Path('template2_target')
    fixtures['template3'] = Path('template3')
    fixtures['template4'] = Path('template4')

    application_config = {
        'module/A': {
            'on_startup': {
                'compile': [
                    {
                        'template': str(fixtures['template1']),
                    },
                    {
                        'template': str(fixtures['template2']),
                        'target': str(fixtures['template2_target']),
                    },
                ],
            },
        },
        'module/B': {
            'on_modified': {
                str(fixtures['template3']): {
                    'run': ['echo template3 touched'],
                    'compile': [
                        {
                            'template': str(fixtures['template4']),
                        },
                    ],
                },
            },
        },
    }
    application_config.update(default_global_options)
    application_config.update(_runtime)
    fixtures['application_config'] = application_config

    module_manager = ModuleManager(application_config)
    fixtures['module_manager'] = module_manager

    yield fixtures

    if fixtures['template2_target'].is_file():
        os.remove(fixtures['template2_target'])

    module_manager.exit()


def test_that_module_manager_detects_all_managed_templates(
    compiling_fixtures,
    test_config_directory,
):
    """Test that all compiled templates are identified."""

    module_manager = compiling_fixtures['module_manager']
    template1 = compiling_fixtures['template1']
    template2 = compiling_fixtures['template2']
    template4 = compiling_fixtures['template4']

    template2_target = compiling_fixtures['template2_target']

    expected_templates = {
        str(template1): test_config_directory / template1,
        str(template2): test_config_directory / template2,
        str(template4): test_config_directory / template4,
    }
    for specified_path, absolute_path in expected_templates.items():
        assert module_manager.templates[specified_path].source == absolute_path

    assert module_manager.templates[str(template2)].target == \
        test_config_directory / template2_target

def test_module_manager_placeholders_are_correctly_generated(
    compiling_fixtures,
    test_config_directory,
):
    """Test that placeholders are generated."""
    module_manager = compiling_fixtures['module_manager']

    template2 = compiling_fixtures['template2']
    template2_target = compiling_fixtures['template2_target']

    string_replacements = module_manager.string_replacements

    assert string_replacements[str(template2)] == str(
        test_config_directory / template2_target,
    )

def test_that_all_modified_observed_files_are_identified(
    compiling_fixtures,
    test_config_directory,
):
    """Test that all files watched for modifications are correctly identified."""
    module_manager = compiling_fixtures['module_manager']
    specified_path = compiling_fixtures['template3']
    path = test_config_directory / specified_path
    module = module_manager.modules['B']

    assert module_manager.on_modified_paths == {
        path: WatchedFile(
            path=path,
            module=module,
            specified_path=str(specified_path),
        )
    }



"""Module for all event_listener classes, keeping track of certain events for modules."""

import abc
from collections import namedtuple
import logging
import time
from datetime import datetime, timedelta
from math import inf
from typing import Dict, Tuple, Union

import pytz
from astral import Location


EventListenerConfig = Dict[str, Union[str, int, float, None]]
logger = logging.getLogger('astrality')


class EventListener(abc.ABC):
    """Class which defines different events."""

    events: Tuple[str, ...]
    default_event_listener_config: EventListenerConfig

    def __init__(self, event_listener_config: EventListenerConfig) -> None:
        """Initialize an event listener based on the configuration of the user."""
        self.name = self.__class__.__name__.lower()

        # Use default values for event listener configuration options that are not
        # specified
        self.event_listener_config = self.default_event_listener_config.copy()
        self.event_listener_config.update(event_listener_config)

    def event(self) -> str:
        """
        Return the current determined event.

        If the event_listener option `force_event` is set, this value will always be
        returned instead of the correct event.
        """
        if self.event_listener_config.get('force_event', False):
            force_event = self.event_listener_config['force_event']

            if not force_event in self.events:
                logger.warning(
                    f'[event_listener/{self.name}] option `force_event` set to '
                    f'{force_event}, which is not a valid event type for '
                    f'the event_listener type "{self.name}": {self.events}. Still using'
                    ' the option in case it is intentional.'
                )

            return force_event  # type: ignore

        return self._event()


    @abc.abstractmethod
    def _event(self) -> str:
        """Return the current determined event."""
        pass

    @abc.abstractmethod
    def time_until_next_event(self) -> timedelta:
        """Return the time remaining until the next event in seconds."""
        pass


class Solar(EventListener):
    """
    EventListener subclass which keeps track of the suns position in the sky.

    It changes event after dawn, sunrise, morning, afternoon, sunset, dusk.
    """
    events = ('sunrise', 'morning', 'afternoon', 'sunset', 'night')
    default_event_listener_config = {
        'type': 'solar',
        'longitude': 0,
        'latitude': 0,
        'elevation': 0,
    }

    def __init__(self, event_listener_config: EventListenerConfig) -> None:
        super().__init__(event_listener_config)
        self.location = self.construct_astral_location()

    def _event(self) -> str:
        now = self.now()

        if now < self.location.sun()['dawn']:
            event = 'night'
        elif now < self.location.sun()['sunrise']:
            event = 'sunrise'
        elif now < self.location.sun()['noon']:
            event = 'morning'
        elif now < self.location.sun()['sunset']:
            event = 'afternoon'
        elif now < self.location.sun()['dusk']:
            event = 'sunset'
        else:
            event = 'night'

        return event

    def time_until_next_event(self) -> timedelta:
        now = self.now()
        try:
            next_event = min(
                utc_time
                for utc_time
                in self.location.sun().values()
                if now < utc_time
            )
        except ValueError as exception:
            if str(exception) == 'min() arg is an empty sequence':
                # None of the solar events this current day are in the future,
                # so we need to compare with solar events tomorrow instead.
                tomorrow = now + timedelta(days=1, seconds=-1)
                next_event = min(
                    utc_time
                    for utc_time
                    in self.location.sun(tomorrow).values()
                    if now < utc_time
                )
            else:
                raise RuntimeError('Could not find the time of the next event')

        return next_event - now

    def now(self) -> datetime:
        """Return the current UTC time."""
        timezone = pytz.timezone('UTC')
        return timezone.localize(datetime.utcnow())

    def construct_astral_location(
        self,
    ) -> Location:
        # Initialize a custom location for astral, as it doesn't necessarily include
        # your current city of residence
        location = Location()

        # These two doesn't really matter
        location.name = 'CityNotImportant'
        location.region = 'RegionIsNotImportantEither'

        # But these are important, and should be provided by the user
        location.latitude = self.event_listener_config['latitude']
        location.longitude = self.event_listener_config['longitude']
        location.elevation = self.event_listener_config['elevation']
        location.timezone = 'UTC'

        return location


class Weekday(EventListener):
    """EventListener subclass which keeps track of the weekdays."""

    events = (
        'monday',
        'tuesday',
        'wednesday',
        'thursday',
        'friday',
        'saturday',
        'sunday',
    )
    default_event_listener_config = {'type': 'weekday'}

    weekdays = dict(zip(range(0,7), events))

    @classmethod
    def _event(cls) -> str:
        """Return the current determined event."""
        return cls.weekdays[datetime.today().weekday()]

    def time_until_next_event(self) -> timedelta:
        """Return the time remaining until the next event in seconds."""
        tomorrow = datetime.today() + timedelta(days=1)
        tomorrow = tomorrow.replace(hour=0, minute=0)
        return tomorrow - datetime.now()


class Periodic(EventListener):
    """Constant frequency EventListener subclass."""

    class Events(tuple):
        def __contains__(self, item) -> bool:
            try:
                if float(item) - int(item) != 0:
                    return False
                else:
                    return 0 <= int(item) < inf
            except ValueError:
                return False

    events = Events()
    default_event_listener_config = {
        'type': 'periodic',
        'seconds': 0,
        'minutes': 0,
        'hours': 0,
        'days': 0,
    }

    def __init__(self, event_listener_config: EventListenerConfig) -> None:
        """Initialize a constant frequency event listener."""

        super().__init__(event_listener_config)

        # Period specified by the user
        self.timedelta = timedelta(  # type: ignore
            seconds=self.event_listener_config['seconds'],
            minutes=self.event_listener_config['minutes'],
            hours=self.event_listener_config['hours'],
            days=self.event_listener_config['days'],
        )

        if self.timedelta.total_seconds() == 0.0:
            # If no period is specified by the user, then 1 hour is used
            self.timedelta = timedelta(hours=1)

        self.initialization_time = datetime.now()

    def _event(self) -> str:
        return str(int(
            (datetime.now() - self.initialization_time) / self.timedelta
        ))

    def time_until_next_event(self) -> timedelta:
        """Return the time remaining until the next period in seconds."""
        return self.timedelta - (datetime.now() - self.initialization_time) % self.timedelta


WorkDay = namedtuple('WorkDay', ('start', 'end',))

class TimeOfDay(EventListener):
    """
    EventListener which keeps track of time intervals each day of the week.

    The variable names used in the implementation, assumes that the 'on'
    events are referring to worktime.
    """

    events = ('on', 'off',)
    default_event_listener_config = {
        'type': 'time_of_day',
        'monday': '09:00-17:00',
        'tuesday': '09:00-17:00',
        'wednesday': '09:00-17:00',
        'thursday': '09:00-17:00',
        'friday': '09:00-17:00',
        'saturday': '',
        'sunday': '',
    }

    def __init__(self, event_listener_config: EventListenerConfig) -> None:
        super().__init__(event_listener_config)
        self.weekdays = (
            'monday',
            'tuesday',
            'wednesday',
            'thursday',
            'friday',
            'saturday',
            'sunday',
        )
        self.workdays: Dict[str, WorkDay] = {}

        for weekday_num, weekday_name in enumerate(self.weekdays):
            work_period = self.event_listener_config[weekday_name]
            if not work_period:
                continue

            work_start, work_end = work_period.split('-')  # type: ignore
            workday = WorkDay(
                start=time.strptime(work_start, '%H:%M'),
                end=time.strptime(work_end, '%H:%M'),
            )
            self.workdays[weekday_name] = workday

    def _event(self) -> str:
        """Return the current determined period."""
        weekday_name = Weekday({'type': 'weekday'}).event()
        if weekday_name not in self.workdays:
            return 'off'
        else:
            now = datetime.now()
            now_hour = now.hour
            now_minute = now.minute

            after_work_start = \
                now_hour >= self.workdays[weekday_name].start.tm_hour and \
                now_minute >= self.workdays[weekday_name].start.tm_min

            before_work_end = \
                now_hour <= self.workdays[weekday_name].end.tm_hour and \
                now_minute <= self.workdays[weekday_name].end.tm_min

            if after_work_start and before_work_end:
                return 'on'
            else:
                return 'off'


    def time_until_next_event(self) -> timedelta:
        """Return the time remaining until the next event in seconds."""
        weekday_name = Weekday({'type': 'weekday'}).event()

        now = datetime.now()
        now_hour = now.hour
        now_minute = now.minute

        if self._event() == 'on':
            # We are within work hours, and it is easy to find the work end for
            # the same day.
            return timedelta(
                hours=self.workdays[weekday_name].end.tm_hour - now_hour,
                minutes=self.workdays[weekday_name].end.tm_min - now_minute + 1,
            )

        else:
            # TODO: This is bad code, and should be done more cleanly in the
            #       future. But it works, and that is good enough for now.

            # The current zero-indexed weekday
            weekday_num = self.weekdays.index(weekday_name)

            # Get the zero indexed workdays
            workday_nums = tuple(
                self.weekdays.index(workday)
                for workday
                in self.workdays
            )

            # Retrieve all zero-indexed workdays that are later in this week
            next_workday_nums = tuple(
                workday_num
                for workday_num
                in workday_nums
                if workday_num > weekday_num
            )

            if len(next_workday_nums) > 0:
                days_until_next_workday = next_workday_nums[0] - weekday_num
                next_workday = self.workdays[self.weekdays[next_workday_nums[0]]]
            else:
                # There are no workdays later in this week, so we need to
                # get the *first* workday from next week instead.
                min_workday_num = min(workday_nums)
                days_until_next_workday = 7 + min_workday_num - weekday_num
                next_workday = self.workdays[self.weekdays[min_workday_num]]

            return timedelta(
                days=days_until_next_workday,
                hours=next_workday.start.tm_hour - now_hour,
                minutes=next_workday.start.tm_min - now_minute + 1,
            )


class Static(EventListener):
    """EventListener subclass which never changes event."""

    events = ('static',)
    default_event_listener_config = {'type': 'static'}

    def _event(self) -> str:
        """Static event listener asways returns the event 'static'."""

        return 'static'

    def time_until_next_event(self) -> timedelta:
        """Returns a 100 year timedelta as an infinite approximation."""

        return timedelta(days=36500)


EVENT_LISTENERS = {
        'periodic': Periodic,
        'solar': Solar,
        'static': Static,
        'time_of_day': TimeOfDay,
        'weekday': Weekday,
}

def event_listener_factory(event_listener_config: Dict[str, Union[str, int]]) -> EventListener:
    event_listener_type = event_listener_config['type']
    return EVENT_LISTENERS[event_listener_type](event_listener_config)  # type: ignore



import os

import pytest

from astrality.module import ModuleManager


@pytest.yield_fixture
def test_target(temp_directory):
    test_target = temp_directory / 'test_target.temp'

    yield  test_target

    if test_target.is_file():
        os.remove(test_target)

def test_that_all_exit_actions_are_correctly_performed(
    default_global_options,
    _runtime,
    test_config_directory,
    test_target,
):
    application_config = {
        'module/car': {
            'on_startup': {
                'import_context': {
                    'from_path': 'context/mercedes.yml',
                },
                'compile': {
                    'template': 'templates/a_car.template',
                    'target': str(test_target),
                },
            },
            'on_exit': {
                'import_context': {
                    'from_path': 'context/tesla.yml',
                },
                'compile': {
                    'template': 'templates/a_car.template',
                    'target': str(test_target),
                },
            },
        },
    }
    application_config.update(default_global_options)
    application_config.update(_runtime)

    module_manager = ModuleManager(application_config)

    # Before we start, the template target should not exist
    assert not test_target.is_file()

    # We finish tasks, reslulting in Mercedes being compiled
    module_manager.finish_tasks()
    with open(test_target) as file:
        assert file.read() == 'My car is a Mercedes'

    # We now exit, and check if the context import and compilation has been
    # performed
    module_manager.exit()
    with open(test_target) as file:
        assert file.read() == 'My car is a Tesla'



"""Tests for the static event listener subclass."""
from datetime import timedelta

from astrality.event_listener import Static


def test_static_events():
    static = Static({})
    assert 'static' in static.events


def test_static_event_listener_until_next_event():
    default_static = Static({})
    assert default_static.time_until_next_event() == timedelta(days=36500)



import os
import shutil
import time
from pathlib import Path

import pytest

from astrality.filewatcher import DirectoryWatcher


@pytest.yield_fixture
def test_files():
    """Return paths related to two test files and cleanup afterwards."""
    watched_directory = Path('/tmp/astrality')
    test_file1 = watched_directory / 'tmp_test_file1'

    recursive_dir = watched_directory / 'test_folder'
    test_file2 = recursive_dir / 'tmp_test_file2'

    yield watched_directory, recursive_dir, test_file1, test_file2

    # Cleanup files
    if test_file1.is_file():
        os.remove(test_file1)
    if test_file2.is_file():
        os.remove(test_file2)
    if recursive_dir.is_dir():
        shutil.rmtree(recursive_dir)


@pytest.yield_fixture
def watch_dir():
    """Instanciate a directory watcher and stop it after its use."""
    class EventSaver:
        """Mock class for testing callback function."""

        def __init__(self):
            self.called = 0

        def save_argument(self, path: Path) -> None:
            self.called += 1
            self.argument = path

    event_saver = EventSaver()

    # Watch a temporary directory
    watched_directory = Path('/tmp/astrality')
    dir_watcher = DirectoryWatcher(
        directory=watched_directory,
        on_modified=event_saver.save_argument,
    )

    yield dir_watcher, event_saver

    dir_watcher.stop()


@pytest.mark.slow
def test_filesystem_watcher(test_files, watch_dir):
    """
    Test correct callback invocation on directory watching.

    Sometimes the on_modified function is called several times by watchdog,
    for a unknown reason. It might be other tests which interfer. We therefore
    check if the lower bound of calls is satisfied, but do not test the exact
    number of calls to on_modified.
    """
    watched_directory, recursive_dir, test_file1, test_file2 = test_files
    dir_watcher, event_saver = watch_dir

    # Start watching the directory
    dir_watcher.start()

    # Nothing has been modified yet
    assert not hasattr(event_saver, 'argument')
    assert event_saver.called == 0

    # Create an empty file
    test_file1.touch()

    # New files are not considered "modified"
    time.sleep(0.7)
    assert not hasattr(event_saver, 'argument')
    assert event_saver.called == 0

    # But when we write to it, it is considered "modified"
    with open(test_file1, 'w') as file:
        file.write('test_content')

    time.sleep(0.7)
    assert event_saver.argument == test_file1
    assert event_saver.called >= 1

    # Create a directory in the watched directory
    recursive_dir.mkdir(parents=True)

    # Subdirectories are not of interest
    time.sleep(0.7)
    assert event_saver.argument == test_file1
    assert event_saver.called >= 1

    # Create a file in the subdirectory
    test_file2.write_text('test')

    # Both the touch event and the write event are considered of interest
    time.sleep(0.7)
    assert event_saver.argument == test_file2
    assert event_saver.called >= 2



class AstralityConfigurationError(BaseException):
    pass

class NonExistentEnabledModule(AstralityConfigurationError):
    pass

class MisconfiguredConfigurationFile(AstralityConfigurationError):
    pass

class GithubModuleError(AstralityConfigurationError):
    pass



"""Tests for string interpolations in ModuleManager class."""

from astrality.module import ModuleManager

def test_use_of_string_interpolations_of_module(
    default_global_options,
    _runtime,
):
    application_config = {
        'module/A': {
            'on_startup': {
                'compile': [
                    {'template': 'some/template'}
                ],
            },
        },
        'module/B': {
            'on_modified': {
                '/what/ever': {
                    'compile': [
                        {
                            'template': '/another/template',
                            'target': '/target/file',
                        },
                    ],
                },
            }
        },
    }
    application_config.update(default_global_options)
    application_config.update(_runtime)

    module_manager = ModuleManager(application_config)

    target1 = str(module_manager.templates['some/template'].target)
    target2 = '/target/file'

    string = 'one: {some/template} two: {/another/template}'
    interpolated_string = 'one: ' + target1 + ' two: ' + target2

    assert module_manager.interpolate_string(string) == interpolated_string



"""Test module for all behaviour related to the import_context action."""
import copy

from astrality.module import ModuleManager

def test_importing_all_context_sections_from_file(
    test_config_directory,
    default_global_options,
    _runtime,
):
    context_file = test_config_directory / 'context' / 'several_sections.yml'
    original_context = {
        'context/section2': {
            'k2_1': 'original_v2_1',
            'k2_2': 'original_v2_2',
        },
        'context/section3': {
            'k3_1': 'original_v3_1',
            'k3_2': 'original_v3_2',
        },
    }

    application_config = {
        'module/A': {
            'on_startup': {
                'import_context': [
                    {'from_path': str(context_file)}
                ]
            },
        },
    }
    application_config.update(copy.deepcopy(original_context))
    application_config.update(default_global_options)
    application_config.update(_runtime)

    module_manager = ModuleManager(application_config)

    # Assert that contents are equal, requiring some trickery, since the
    # application_context has been processed.
    assert len(module_manager.application_context) == len(original_context)
    for key, value in module_manager.application_context.items():
        assert value == original_context['context/' + key]

    # We expect the following context after import
    expected_context = {
        'context/section1': {
            'k1_1': 'v1_1',
            'k1_2': 'v1_2',
        },
        'context/section2': {
            'k2_1': 'v2_1',
            'k2_2': 'v2_2',
        },
        'context/section3': {
            'k3_1': 'original_v3_1',
            'k3_2': 'original_v3_2',
        },
    }

    # Now run startup actions, resulting in the file being imported
    module_manager.finish_tasks()

    # Assert that the new application_context is as expected
    assert len(module_manager.application_context) == len(expected_context)
    for key, value in module_manager.application_context.items():
        assert value == expected_context['context/' + key]



"""Specifies everything related to application spanning configuration."""

import copy
import logging
import os
import re
from abc import ABC, abstractclassmethod, abstractmethod
from distutils.dir_util import copy_tree
from io import StringIO
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Match,
    MutableMapping,
    Optional,
    Pattern,
    Tuple,
    Type,
    Iterable,
    Union,
)
import re

from mypy_extensions import TypedDict

from astrality import compiler
from astrality.exceptions import (
    MisconfiguredConfigurationFile,
    NonExistentEnabledModule,
)
from astrality.github import clone_repo, clone_or_pull_repo
from astrality.resolver import Resolver
from astrality.utils import generate_expanded_env_dict, run_shell

Context = Dict[str, Resolver]

logger = logging.getLogger('astrality')

from yaml import load, dump
try:
    from yaml import CLoader as Loader, CDumper as Dumper  # type: ignore
    logger.info('Using LibYAML bindings for faster .yml parsing.')
except ImportError:  # pragma: no cover
    from yaml import Loader, Dumper
    logger.warning(
        'LibYAML not installed.'
        'Using somewhat slower pure python implementation.',
    )


ApplicationConfig = Dict[str, Dict[str, Any]]

ASTRALITY_DEFAULT_GLOBAL_SETTINGS = {'config/astrality': {
    'hot_reload_config': False,
    'startup_delay': 0,
}}


def resolve_config_directory() -> Path:
    """
    Return the absolute configuration directory path for the application.

    The directory path is resolved as follows:
        1) If $ASTRALITY_CONFIG_HOME is present, use it.
        2) If $XDG_CONFIG_HOME is present, use $XDG_CONFIG_HOME/astrality.
        3) Elsewise, use ~/config/astrality.
    """
    if 'ASTRALITY_CONFIG_HOME' in os.environ:
        # The user has set a custom config directory for astrality
        config_directory = Path(os.environ['ASTRALITY_CONFIG_HOME'])
    else:
        # Follow the XDG directory standard
        config_directory = Path(
            os.getenv('XDG_CONFIG_HOME', '~/.config'),
            'astrality',
        )
    return config_directory.expanduser().absolute()


def infer_config_location(
    config_directory: Optional[Path] = None,
) -> Tuple[Path, Path]:
    """
    Return path of Astrality configuration file based on config folder path.

    Try to find the configuration directory and file for astrality, based on
    filesystem or specific environment variables if they are present several
    places to put it. If the expected config file is not present, use an
    example configuration instead.
    """
    if not config_directory:
        config_directory = resolve_config_directory()
    config_file = Path(config_directory, 'astrality.yml')

    if not config_file.is_file():
        logger.warning(
            'Configuration file not found in its expected path '
            + str(config_file) +
            '.'
        )
        config_directory = Path(__file__).parent.absolute() / 'config'
        config_file = config_directory / 'astrality.yml'
        logger.warning(f'Using example configuration instead: "{config_file}"')
    else:
        logging.info(f'Using configuration file "{config_file}"')

    return config_directory, config_file


def dict_from_config_file(
    config_file: Path,
    context: Dict[str, Resolver],
) -> ApplicationConfig:
    """
    Return a dictionary that reflects the contents of `config_file`.

    Environment variables are interpolated like this:
        {{ env:NAME_OF_ENV_VARIABLE }} -> os.environ[NAME_OF_ENV_VARIABLE]

    And shell commands can be inserted like this:
        {{ 'shell command' | shell }}
    """

    if not config_file.is_file():  # pragma: no cover
        error_msg = f'Could not load config file "{config_file}".'
        logger.critical(error_msg)
        raise FileNotFoundError(error_msg)

    config_string = compiler.compile_template_to_string(
        template=config_file,
        context=context,
        shell_command_working_directory=config_file.parent,
    )
    conf_dict = load(StringIO(config_string))

    return conf_dict


def infer_runtime_variables_from_config(
    config_directory: Path,
    config_file: Path,
    config: ApplicationConfig,
) -> Dict[str, Dict[str, Path]]:
    """Return infered runtime variables based on config file."""

    temp_directory = Path(os.environ.get('TMPDIR', '/tmp')) / 'astrality'
    if not temp_directory.is_dir():
        os.mkdir(temp_directory)

    return {
        '_runtime': {
            'config_directory': config_directory,
            'config_file': config_file,
            'temp_directory': temp_directory,
        }
    }


def user_configuration(config_directory: Optional[Path] = None) -> ApplicationConfig:
    """
    Return Resolver object containing the users configuration.

    Create a configuration dictionary which should directly reflect the
    hierarchy of a typical `astrality.yml` file. Users should be able to insert
    elements from their configuration directly into conky module templates. The
    mapping should be:

    ${astrality:fonts:1} -> config['fonts']['1']

    In addition, the section config['_runtime'] is inserted, which contains
    several items specifying runtime specific values. Example keys are:
    - config_directory
    - config_file
    - temp_directory
    """
    config_directory, config_file = infer_config_location(config_directory)

    config = dict_from_config_file(config_file=config_file, context={})
    config.update(infer_runtime_variables_from_config(
        config_directory,
        config_file,
        config,
    ))

    # Insert default global settings that are not specified
    user_settings = config.get('config/astrality', {})
    config['config/astrality'] = ASTRALITY_DEFAULT_GLOBAL_SETTINGS['config/astrality'].copy()
    config['config/astrality'].update(user_settings)

    return config


def insert_into(
    context: Context,
    from_config_file: Path,
    section: Optional[str],
    from_section: Optional[str],
) -> Context:
    """
    Import section(s) from config file into config dictionary.

    If `section` and `from_section` are given:
    The method overwrites `config[section]` with the values from [from_section]
    defined in `from_config_file`.

    Else:
    All sections are imported from `from_config_file`.
    """
    logger.info(
        f'Importing context section {section} from {str(from_config_file)}',
    )

    # Context files do not insert context values, as that would cause a lot
    # of complexity for end users. Old context values will be inserted
    # for placeholders, etc.
    contexts = compiler.context(dict_from_config_file(
        from_config_file,
        context={},
    ))

    if section and from_section:
        # A specific source and target section has been specified
        context[section] = contexts[from_section]
    else:
        # All sections should be merged into the context
        context.update(contexts)

    return context


def create_config_directory(path: Optional[Path]=None, empty=False) -> Path:
    """Create application configuration directory and return its path."""
    if not path:
        path = resolve_config_directory()

    if not path.exists():
        if empty:
            logger.warning(f'Creating empty directory at "{str(path)}".')
            path.mkdir(parents=True)
        else:
            logger.warning(f'Copying over example config directory to "{str(path)}".')
            example_config_dir = Path(__file__).parent / 'config'
            copy_tree(
                src=str(example_config_dir),
                dst=str(path),
            )
    else:  # pragma: no cover
        logger.warning(f'Path "{str(path)}" already exists! Delete it first.')

    return path

def expand_path(path: Path, config_directory: Path) -> Path:
    """
    Return an absolute path from a (possibly) relative path.

    Relative paths are relative to $ASTRALITY_CONFIG_HOME, and ~ is
    expanded to the home directory of $USER.
    """

    path = Path.expanduser(path)

    if not path.is_absolute():
        path = Path(
            config_directory,
            path,
        )

    return path


class EnablingStatementRequired(TypedDict):
    name: str


class EnablingStatement(EnablingStatementRequired, total=False):
    """Dictionary defining an externally defined module."""
    trusted: bool
    autoupdate: bool


ModuleConfig = Dict[str, Any]


class ModuleSource(ABC):
    directory: Path
    config_file: Path
    _config: Dict[Any, Any]

    @abstractmethod
    def __init__(
        self,
        enabling_statement: EnablingStatement,
        modules_directory: Path,
    ) -> None:
        """Initialize a module source from an enabling statement."""
        raise NotImplementedError

    @property
    @abstractmethod
    def name_syntax(self) -> Pattern:
        """Regular expression defining how to name this type of module source."""
        raise NotImplementedError

    @abstractmethod
    def config(self, context: Dict[str, Resolver]) -> Dict[Any, Any]:
        """Return the dictionary containing the module configuration."""
        raise NotImplementedError

    @classmethod
    def represented_by(cls, module_name: str) -> bool:
        """Return True if name represents module source type."""
        return bool(cls.name_syntax.match(module_name))

    @classmethod
    def type(cls, of: str) -> 'ModuleSource':
        """
        Return the source subclass which is responsible for the module name.
        """
        for source_type in cls.__subclasses__():
            if source_type.represented_by(module_name=of):
                return source_type

        raise MisconfiguredConfigurationFile(  # pragma: no cover
            f'Tried to enable invalid module name syntax "{of}".',
        )

    @abstractmethod
    def __contains__(self, module_name: str) -> bool:
        """Return True module source contains module with name `module_name`."""
        raise NotImplementedError


class GlobalModuleSource(ModuleSource):
    """Module defined in `astrality.yml`."""
    name_syntax = re.compile(r'^(\w+|\*)$')

    def __init__(
        self,
        enabling_statement: EnablingStatement,
        modules_directory: Path,
    ) -> None:
        """Initialize enabled module defined in `astrality.yml`."""
        self.enabled_module = enabling_statement['name']
        assert self.represented_by(module_name=self.enabled_module)

        assert modules_directory.is_absolute()
        self.directory = modules_directory

    def config(self, context: Dict[str, Resolver]) -> Dict[Any, Any]:
        """Return configuration related to global module (TODO)."""
        return {}

    def __contains__(self, module_name: str) -> bool:
        """Return True if the module name is enabled."""
        if self.enabled_module == '*' and self.represented_by(module_name):
            return True
        else:
            return module_name == self.enabled_module

    def __repr__(self) -> str:
        return f"GlobalModuleSource('{self.enabled_module}')"


class GithubModuleSource(ModuleSource):
    """Module defined in a GitHub repository."""
    name_syntax = re.compile(r'^github::.+/.+(::(\w+|\*))?$')
    _config: ApplicationConfig

    def __init__(
        self,
        enabling_statement: EnablingStatement,
        modules_directory: Path,
    ) -> None:
        """
        Initialize enabled module defined in Github repository

        A GitHub module is enabled with the name syntax:
        github::<github_user>/<github_repo>[::<enabled_module>]

        If <enabled_module> is skipped, all modules are enabled, i.e.
        github::<github_user>/<github_repo>::*
        """
        assert self.represented_by(
            module_name=enabling_statement['name'],
        )
        specified_module = enabling_statement['name']
        self.autoupdate = enabling_statement.get('autoupdate', False)

        github_path, *enabled_modules = specified_module[8:].split('::')

        if len(enabled_modules) > 0:
            self.enabled_module_name = enabled_modules[0]
        else:
            # Enable all modules since all have been enabled
            self.enabled_module_name = '*'

        self.github_user, self.github_repo = github_path.split('/')

        assert modules_directory.is_absolute()
        self.modules_directory = modules_directory
        self.directory = modules_directory / self.github_user / self.github_repo
        self.config_file = self.directory / 'config.yml'

    def config(self, context: Dict[str, Resolver]) -> Dict[Any, Any]:
        """Return the contents of config.yml, filtering out disabled modules."""
        if hasattr(self, '_config'):
            return self._config

        # The repository already exists, so it is assumed that the user wants
        # the repository cloned
        if not self.directory.is_dir():
            clone_repo(
                user=self.github_user,
                repository=self.github_repo,
                modules_directory=self.modules_directory,
            )
        elif self.autoupdate:
            clone_or_pull_repo(
                user=self.github_user,
                repository=self.github_repo,
                modules_directory=self.modules_directory,
            )

        self._config = filter_config_file(
            config_file=self.config_file,
            context=context,
            enabled_module_name=self.enabled_module_name,
            prepend=f'github::{self.github_user}/{self.github_repo}::',
        )

        return self._config

    def __contains__(self, module_name: str) -> bool:
        """Return True if GitHub module repo is enabled."""
        return 'module/' + module_name in self._config

    def __eq__(self, other) -> bool:
        """
        Return True if two GithubModuleSource objects represent the same
        directory with identical modules enabled.
        """
        try:
            return self.directory == other.directory and self._config == other._config
        except AttributeError:
            return False



    def __repr__(self) -> str:
        return f"GithubModuleSource('{self.github_user}/{self.github_repo}')" \
            f' = {tuple(self._config.keys())}'


class DirectoryModuleSource(ModuleSource):
    """
    Module(s) defined in a directory.

    Specifically: `$ASTRALITY_CONFIG_HOME/{modules_directory}/config.yml
    """

    name_syntax = re.compile(r'^(?!github::).+::(\w+|\*)$')

    def __init__(
        self,
        enabling_statement: EnablingStatement,
        modules_directory: Path,
    ) -> None:
        """
        Initialize an ExternalModule object.
        """
        assert self.represented_by(
            module_name=enabling_statement['name'],
        )
        relative_directory_path, self.enabled_module_name = enabling_statement['name'].split('::')
        self.relative_directory_path = Path(relative_directory_path)

        assert modules_directory.is_absolute()
        self.directory = modules_directory / self.relative_directory_path

        self.config_file = self.directory / 'config.yml'
        self.trusted = enabling_statement.get('trusted', True)

    def config(self, context: Dict[str, Resolver]) -> Dict[Any, Any]:
        """Return the module configuration defined in directory."""
        if not hasattr(self, '_config'):
            self._config = filter_config_file(
                config_file=self.config_file,
                context=context,
                enabled_module_name=self.enabled_module_name,
                prepend=str(self.relative_directory_path) + '::',
            )

        return self._config

    def __repr__(self):
        """Human-readable representation of a DirectoryModuleSource object."""
        return f'DirectoryModuleSource(name={self.relative_directory_path}::{self.enabled_module_name}, directory={self.directory}, trusted={self.trusted})'

    def __eq__(self, other) -> bool:
        """
        Return true if two DirectoryModuleSource objects represents the same Module.

        Entirily determined by the source directory of the module.
        """
        try:
            return self.directory == other.directory
        except AttributeError:  # pragma: no cover
            return False

    def __contains__(self, module_name: str) -> bool:
        """Return True if source contains module named `module_name`."""
        return 'module/' + module_name in self._config


class EnabledModules:
    """
    Object which keeps track of all enabled modules.

    Also allows fetching of configurations defined externally.
    """

    def __init__(
        self,
        enabling_statements: List[EnablingStatement],
        config_directory: Path,
        modules_directory: Path,
    ) -> None:
        """Determine exactly which modules are enabled from user config."""
        enabling_statements = self.process_enabling_statements(
            enabling_statements=enabling_statements,
            modules_directory=modules_directory,
        )

        self.source_types: Dict[Type[ModuleSource], List[ModuleSource]] = {
            GlobalModuleSource: [],
            DirectoryModuleSource: [],
            GithubModuleSource: [],
        }

        for enabling_statement in enabling_statements:
            try:
                source_type = ModuleSource.type(of=enabling_statement['name'])
            except MisconfiguredConfigurationFile:
                logger.error(
                    f'Invalid module name syntax {str(enabling_statement)} '
                    'in enabled_modules configuration.'
                )
                continue

            if source_type == GlobalModuleSource:
                source_directory = config_directory
            else:
                source_directory = modules_directory

            self.source_types[source_type].append(source_type(
                enabling_statement=enabling_statement,
                modules_directory=source_directory,
            ))

    def process_enabling_statements(
        self,
        enabling_statements: List[EnablingStatement],
        modules_directory: Path,
    ):
        """
        Return enabling statements where wildcards have been replaced.
        """
        self.all_global_modules_enabled = False
        self.all_directory_modules_enabled = False

        new_enabling_statements: List[EnablingStatement] = []
        for enabling_statement in enabling_statements:
            if enabling_statement['name'] == '*':
                self.all_global_modules_enabled = True
                new_enabling_statements.append(enabling_statement)

            elif enabling_statement['name'] == '*::*':
                self.all_directory_modules_enabled = True

                for module_directory in self.module_directories(within=modules_directory):
                    directory_module = module_directory + '::*'
                    new_enabling_statement = copy.deepcopy(enabling_statement)
                    new_enabling_statement['name'] = directory_module
                    new_enabling_statements.append(new_enabling_statement)

            else:
                new_enabling_statements.append(enabling_statement)

        return new_enabling_statements

    @staticmethod
    def module_directories(within: Path) -> Tuple[str, ...]:
        try:
            return tuple(
                path.name
                for path
                in within.iterdir()
                if (path / 'config.yml').is_file()
            )
        except FileNotFoundError:
            logger.error(
                f'Tried to search for module directories in "{within}", '
                'but directory does not exist!.'
            )
            return ()

    def compile_config_files(
        self,
        context: Dict[str, Resolver],
    ):
        """Compile all config templates with context."""
        for source in (
            *self.source_types[DirectoryModuleSource],
            *self.source_types[GithubModuleSource],
        ):
            source.config(context=context)


    def __contains__(self, module_name: str) -> bool:
        """Return True if the given module name is supposed to be enabled."""
        if module_name[:7].lower() == 'module/':
            module_name = module_name[7:]

        source_type = ModuleSource.type(of=module_name)
        for module_source in self.source_types[source_type]:
            if module_name in module_source:
                return True

        return False

    def __repr__(self) -> str:
        return ', '.join(
            source.__repr__()
            for source
            in self.source_types.values()
        )


class GlobalModulesConfigDict(TypedDict, total=False):
    """Dictionary defining configuration options for Modules."""
    requires_timeout: Union[int, float]
    run_timeout: Union[int, float]
    recompile_modified_templates: bool
    modules_directory: str
    enabled_modules: List[EnablingStatement]


class GlobalModulesConfig:
    """
    User modules configuration.

    The plan is to make this the input argument for ModuleManager.__init__(),
    extracting all logic related to getting module configurations (only enebled
    ones) from the object.
    """

    def __init__(
        self,
        config: GlobalModulesConfigDict,
        config_directory: Path,
    ) -> None:
        """Initialize a GlobalModulesConfig object from a dictionary."""

        self.recompile_modified_templates = config.get(
            'recompile_modified_templates',
            False,
        )
        self.requires_timeout = config.get(
            'requires_timeout',
            1,
        )
        self.run_timeout = config.get(
            'run_timeout',
            0,
        )

        # Determine the directory which contains external modules
        assert config_directory.is_absolute()
        self.config_directory = config_directory
        if 'modules_directory' in config:
            # Custom modules folder
            modules_path = expand_path(
                path=Path(config['modules_directory']),
                config_directory=self.config_directory,
            )
            self.modules_directory = modules_path
        else:
            # Default modules folder: $ASTRALITY_CONFIG_HOME/modules
            self.modules_directory = config_directory / 'modules'

        # Enable all modules if nothing is specified
        self.enabled_modules = EnabledModules(
            enabling_statements=config.get(
                'enabled_modules',
                [
                    {'name': '*', 'trusted': True},
                    {'name': '*::*', 'trusted': True},
                ]
            ),
            config_directory=self.config_directory,
            modules_directory=self.modules_directory,
        )

    @property
    def external_module_sources(
        self,
    ) -> Iterable[ModuleSource]:
        """Return an iterator yielding all external module configs."""
        for source in (
            *self.enabled_modules.source_types[DirectoryModuleSource],
            *self.enabled_modules.source_types[GithubModuleSource],
        ):
            yield source

    @property
    def external_module_config_files(self) -> Iterable[Path]:
        """Return an iterator yielding all absolute paths to module config files."""
        for external_module_source in self.external_module_sources:
            yield external_module_source.config_file

    def compile_config_files(
        self,
        context: Dict[str, Resolver],
    ):
        """Compile all config templates with context."""
        self.enabled_modules.compile_config_files(context)


def filter_config_file(
    config_file: Path,
    context: Dict[str, Resolver],
    enabled_module_name: str,
    prepend: str,
) -> ModuleConfig:
    """
    Return a filtered dictionary representing `config_file`.

    Only modules given by `enabled_module_name` are kept, and their names
    are prepended with `prepend`.
    """
    try:
        modules_dict = dict_from_config_file(
            config_file=config_file,
            context=context,
        )
    except FileNotFoundError:
        logger.warning(
            f'Non-existent module configuration file "{config_file}"'
            'Skipping enabled module '
            f'"{prepend}{enabled_module_name}"'
        )
        return {}

    if not modules_dict:  # pragma: no cover
        logger.warning(
            f'Empty modules configuration "{config_file}".',
        )
        return {}
    elif not isinstance(modules_dict, dict):  # pragma: no cover
        logger.critical(
            f'Configuration file "{config_file}" not formated as '
            'a dictionary at root indentation.'
        )
        raise MisconfiguredConfigurationFile

    sections = tuple(modules_dict.keys())
    if enabled_module_name != '*' \
            and 'module/' + enabled_module_name not in sections:
        raise NonExistentEnabledModule

    # We rename each module to module/{self.name}.module_name
    # in order to prevent naming conflicts when using modules provided
    # from a third party with the same name as another managed module.
    # This way you can use a module named "conky" from two third parties,
    # in addition to providing your own.
    for section in sections:
        if len(section) > 7 and section[:7].lower() == 'module/':
            # This section defines a module

            module_name = section[7:]
            if not enabled_module_name == '*' \
               and enabled_module_name != module_name:
                # The module is not enabled, remove the module
                modules_dict.pop(section)
                continue

            # Replace the module name with folder_name.module_name
            non_conflicting_module_name = \
                'module/' + prepend + module_name
            module_section = modules_dict.pop(section)
            modules_dict[non_conflicting_module_name] = module_section

    return modules_dict



from datetime import datetime, timedelta

import pytest

from astrality.event_listener import Periodic


@pytest.fixture
def periodic():
    """Return a default periodic timer."""
    return Periodic({'type': 'periodic'})


@pytest.mark.parametrize("event, is_event", [
    ('0', True),
    ('1', True),
    ('80', True),
    ('-1', False),
    ('1.2', False),
    ('night', False),
])
def test_periodic_events_are_correctly_identified(event, is_event, periodic):
    """Test that events are correctly identified as valid."""
    assert (event in periodic.events) == is_event


def test_periodic_beginning_event(freezer):
    """Test that the first event is 0."""
    periodic = Periodic({})
    assert periodic.event() == '0'


def test_periodic_standard_timedelta(freezer):
    """Test that the maximum timedelta is returned."""
    default_periodic = Periodic({})
    assert default_periodic.time_until_next_event().total_seconds() == 3600


def test_using_custom_periodic_event_listener(freezer):
    """Test that a custom defined period is correctly behaved."""
    custom_periodic_config = {
        'type': 'periodic',
        'seconds': 1,
        'minutes': 2,
        'hours': 3,
        'days': 4,
    }
    custom_periodic = Periodic(custom_periodic_config)
    assert custom_periodic.time_until_next_event() == timedelta(
        seconds=1,
        minutes=2,
        hours=3,
        days=4,
    )


def test_weekday_time_until_next_event_of_periodic_event_listener(freezer):
    """Test that time left until next event is always correct."""
    now = datetime.now()
    freezer.move_to(now)
    default_periodic = Periodic({})
    assert default_periodic.time_until_next_event().total_seconds() == 60*60

    fourty_minutes = timedelta(minutes=40)
    freezer.move_to(now + fourty_minutes)
    assert default_periodic.time_until_next_event().total_seconds() == 20*60

    freezer.move_to(now + 2 * fourty_minutes)
    assert default_periodic.time_until_next_event().total_seconds() == 40*60


def test_enumeration_of_periodic_event_listener_events(freezer):
    """Test that event is incremented by one correctly."""
    now = datetime.now()
    freezer.move_to(now)
    default_periodic = Periodic({})
    assert default_periodic.event() == '0'

    fourty_minutes = timedelta(minutes=40)
    freezer.move_to(now + fourty_minutes)
    assert default_periodic.event() == '0'

    freezer.move_to(now + 2 * fourty_minutes)
    assert default_periodic.event() == '1'



import logging
from pathlib import Path

from astrality.module import Module

def test_module_requires_option(caplog):
    """Test that modules are disabled when they don't satisfy `requires`."""

    # Simple module that satisfies the requirements
    does_satisfy_requiremnets = { 'module/satisfies': {
        'enabled': True,
        'requires': 'command -v cd',
    }}
    assert Module.valid_class_section(
        section=does_satisfy_requiremnets,
        requires_timeout=1,
        requires_working_directory=Path('/'),
    )


    # Simple module that does not satisfy requirements
    does_not_satisfy_requirements = { 'module/does_not_satisfy': {
        'requires': 'command -v does_not_exist',
    }}
    assert not Module.valid_class_section(
        section=does_not_satisfy_requirements,
        requires_timeout=1,
        requires_working_directory=Path('/'),
    )
    assert (
        'astrality',
        logging.WARNING,
        '[module/does_not_satisfy] Module does not satisfy requirement "command -v does_not_exist".',
    ) in caplog.record_tuples


    # Test failing one of many requirements
    does_not_satisfy_one_requirement = { 'module/does_not_satisfy': {
        'requires': ['command -v cd', 'command -v does_not_exist'],
    }}
    caplog.clear()
    assert not Module.valid_class_section(
        section=does_not_satisfy_requirements,
        requires_timeout=1,
        requires_working_directory=Path('/'),
    )
    assert (
        'astrality',
        logging.WARNING,
        '[module/does_not_satisfy] Module does not satisfy requirement "command -v does_not_exist".',
    ) in caplog.record_tuples



import os
from datetime import datetime
from pathlib import Path
from shutil import rmtree

import pytest

from astrality import compiler
from astrality.config import (
    create_config_directory,
    dict_from_config_file,
    user_configuration,
    expand_path,
    insert_into,
    resolve_config_directory,
)
from astrality.module import ModuleManager
from astrality.utils import generate_expanded_env_dict


@pytest.fixture
def dummy_config():
    test_conf = Path(__file__).parents[1] / 'test_config' / 'test.yml'
    return dict_from_config_file(
        config_file=test_conf,
        context={},
    )


class TestAllConfigFeaturesFromDummyConfig:
    def test_normal_variable(self, dummy_config):
        assert dummy_config['context/section1']['var1'] == 'value1'

    def test_variable_interpolation(self, dummy_config):
        assert dummy_config['context/section1']['var2'] == 'value1/value2'
        assert dummy_config['context/section2']['var3'] == 'value1'

    def test_empty_string_variable(self, dummy_config):
        assert dummy_config['context/section2']['empty_string_var'] == ''

    def test_non_existing_variable(self, dummy_config):
        with pytest.raises(KeyError):
            assert dummy_config['context/section2']['not_existing_option'] is None

    def test_environment_variable_interpolation(self, dummy_config):
        assert dummy_config['context/section3']['env_variable'] == 'test_value, hello'


def test_config_directory_name(conf):
    assert str(conf['_runtime']['config_directory'])[-7:] == '/config'


def test_name_of_config_file(conf):
    assert '/astrality.yml' in str(conf['_runtime']['config_file'])


def test_generation_of_expanded_env_dict():
    env_dict = generate_expanded_env_dict()
    assert len(env_dict) == len(os.environ)

    for name, value in os.environ.items():
        if not '$' in value:
            assert env_dict[name] == value

def test_insert_context_section():
    context = compiler.context({'context/section1': {'key_one': 'value_one'}})
    assert context['section1']['key_one'] == 'value_one'

    test_config_file = Path(__file__).parents[1] / 'test_config' / 'test.yml'
    context = insert_into(
        context=context,
        section='new_section',
        from_config_file=test_config_file,
        from_section='section2'
    )

    assert context['section1']['key_one'] == 'value_one'
    assert context['new_section']['var3'] == 'value1'

    context = insert_into(
        context=context,
        section='section3',
        from_config_file=test_config_file,
        from_section='section3'
    )
    assert context['section3']['env_variable'] == 'test_value, hello'

    context = insert_into(
        context=context,
        section='section1',
        from_config_file=test_config_file,
        from_section='section1'
    )
    assert context['section1']['var2'] == 'value1/value2'
    assert 'key_one' not in context['section1']


class TestResolveConfigDirectory:
    def test_setting_directory_using_application_env_variable(self, monkeypatch):
        monkeypatch.setattr(
            os,
            'environ',
            {
                'ASTRALITY_CONFIG_HOME': '/test/dir',
                'XDG_CONFIG_HOME': '/xdg/dir',
            },
        )
        assert resolve_config_directory() == Path('/test/dir')

    def test_setting_directory_using_xdg_directory_standard(self, monkeypatch):
        monkeypatch.setattr(
            os,
            'environ',
            {
                'XDG_CONFIG_HOME': '/xdg/dir',
            },
        )
        assert resolve_config_directory() == Path('/xdg/dir/astrality')

    def test_using_standard_config_dir_when_nothing_else_is_specified(self, monkeypatch):
        monkeypatch.setattr(os, 'environ', {})
        assert resolve_config_directory() == Path('~/.config/astrality').expanduser()


class TestCreateConfigDirectory:
    def test_creation_of_empty_config_directory(self):
        config_path = Path('/tmp/config_test')
        config_dir = create_config_directory(path=config_path, empty=True)
        assert config_path == config_dir
        assert config_dir.is_dir()
        assert len(list(config_dir.iterdir())) == 0
        config_dir.rmdir()

    def test_creation_of_infered_config_directory(self, monkeypatch):
        config_path = Path('/tmp/astrality_config')
        monkeypatch.setattr(
            os,
            'environ',
            {'ASTRALITY_CONFIG_HOME': str(config_path)},
        )
        created_config_dir = create_config_directory(empty=True)
        assert created_config_dir == config_path
        created_config_dir.rmdir()

    def test_creation_of_config_directory_with_example_content(self):
        """Test copying example configuration contents."""
        config_path = Path('/tmp/astrality_config_with_contents')
        created_config_dir = create_config_directory(config_path)
        assert created_config_dir == config_path

        # Test presence of content in created folder
        dir_contents = tuple(file.name for file in created_config_dir.iterdir())
        assert 'astrality.yml' in dir_contents
        assert 'modules' in dir_contents
        rmtree(created_config_dir)

def test_expand_path_method(test_config_directory):
    absolute_path = Path('/tmp/ast')
    tilde_path = Path('~/dir')
    relative_path = Path('test')

    assert expand_path(
        path=absolute_path,
        config_directory=Path('/what/ever'),
    ) == absolute_path

    assert expand_path(
        path=tilde_path,
        config_directory=Path('/what/ever'),
    ) == Path.home() / 'dir'

    assert expand_path(
        path=relative_path,
        config_directory=test_config_directory,
    ) == test_config_directory / 'test'

@pytest.yield_fixture
def dir_with_compilable_files(tmpdir):
    config_dir = Path(tmpdir)
    config_file = config_dir / 'astrality.yml'
    config_file.write_text(
        'key1: {{ env.EXAMPLE_ENV_VARIABLE }}\n'
        'key2: {{ "echo test" | shell }}'
    )

    module_file = config_dir / 'config.yml'
    module_file.write_text(
        'key1: {{ env.EXAMPLE_ENV_VARIABLE }}\n'
        'key2: {{ "echo test" | shell }}'
    )

    yield config_dir

    os.remove(config_file)
    os.remove(module_file)
    config_dir.rmdir()

class TestUsingConfigFilesWithPlaceholders:
    def test_dict_from_config_file(self, dir_with_compilable_files):
        config = dict_from_config_file(
            config_file=dir_with_compilable_files / 'astrality.yml',
            context={},
        )
        assert config == {
            'key1': 'test_value',
            'key2': 'test',
        }

    def test_get_user_configuration(self, dir_with_compilable_files):
        user_conf = user_configuration(dir_with_compilable_files)
        assert user_conf['key1'] == 'test_value'
        assert user_conf['key2'] == 'test'

    def test_module_directory_source(
        self,
        test_config_directory,
        default_global_options,
        _runtime,
    ):
        application_config = {
            'config/modules': {
                'modules_directory': 'test_modules',
                'enabled_modules': [
                    {'name': 'module_with_placeholders::*'},
                ],
            },
            'context/module': {
                'setting': 'hi',
            },
        }
        application_config.update(default_global_options)
        application_config.update(_runtime)

        module_manager = ModuleManager(application_config)
        module_config = module_manager.modules['module_with_placeholders::placeholders'].module_config

        assert module_config['hi'] == 'bye!'
        assert module_config['test_value'] == 'test'



"""Tests for module manager behaviour related to file system modifications."""
import os
import shutil
import time
from pathlib import Path

import pytest

from astrality.config import dict_from_config_file
from astrality.module import ModuleManager


@pytest.yield_fixture
def modules_config(
    test_config_directory,
    default_global_options,
    _runtime,
    temp_directory,
):
    empty_template = test_config_directory / 'templates' / 'empty.template'
    empty_template_target = Path('/tmp/astrality/empty_temp_template')
    touch_target = temp_directory / 'touched'

    secondary_template = test_config_directory / 'templates' / 'no_context.template'
    secondary_template_target = temp_directory / 'secondary_template.tmp'

    config = {
        'module/A': {
            'on_modified': {
                str(empty_template): {
                    'compile': [
                        {
                            'template' : str(empty_template),
                            'target': str(empty_template_target),
                        },
                        {
                            'template': str(secondary_template),
                            'target': str(secondary_template_target),
                        },
                    ],
                    'run': ['touch ' + str(touch_target)],
                },
            },
        },
        'module/B': {},
    }
    config.update(default_global_options)
    config.update(_runtime)
    yield (
        config,
        empty_template,
        empty_template_target,
        touch_target,
        secondary_template,
        secondary_template_target,
    )

    # Cleanup all test files for next test iteration
    if empty_template.is_file():
        empty_template.write_text('')
    if empty_template_target.is_file():
        os.remove(empty_template_target)
    if secondary_template_target.is_file():
        os.remove(secondary_template_target)
    if touch_target.is_file():
        os.remove(touch_target)


def test_modified_commands_of_module(modules_config):
    config, empty_template, empty_template_target, touch_target, *_= modules_config
    module_manager = ModuleManager(config)
    assert module_manager.modules['A'].modified_commands(str(empty_template)) == \
        ('touch ' + str(touch_target), )

def test_direct_invocation_of_modifed_method_of_module_manager(modules_config):
    (
        config,
        empty_template,
        empty_template_target,
        touch_target,
        secondary_template,
        secondary_template_target,
    ) = modules_config
    module_manager = ModuleManager(config)

    # PS: Disabling the directory watcher is not necessary, as it is done in
    # the startup method.

    # Now write new text to the template
    empty_template.write_text('new content')

    # And trigger the modified method manually
    module_manager.file_system_modified(empty_template)

    # And assert that the new template has been compiled
    assert empty_template_target.is_file()
    with open(empty_template_target) as file:
        assert file.read() == 'new content'

    # And that the new file has been touched
    time.sleep(0.5)
    assert touch_target.is_file()

def test_on_modified_event_in_module(modules_config):
    (
        config,
        empty_template,
        empty_template_target,
        touch_target,
        secondary_template,
        secondary_template_target,
    ) = modules_config
    module_manager = ModuleManager(config)

    # Start the file watcher by invoking the startup command indirectly
    # through finish_tasks() method
    module_manager.finish_tasks()

    # Assert that the template file is really empty as a sanity check
    with open(empty_template) as file:
        assert file.read() == ''

    # And that target files are not created yet
    assert not touch_target.is_file()
    assert not empty_template_target.is_file()
    assert not secondary_template_target.is_file()

    # Trigger the on_modified event
    empty_template.write_text('new content')
    time.sleep(1)

    # And assert that the new template has been compiled
    assert empty_template_target.is_file()
    with open(empty_template_target) as file:
        assert file.read() == 'new content'

    # Assert that also templates from other modules are compiled
    assert secondary_template_target.is_file()
    with open(secondary_template_target) as file:
        assert file.read() == 'one\ntwo\nthree'

    # And that the new file has been touched
    assert touch_target.is_file()

@pytest.yield_fixture
def test_template_targets():
    template_target1 = Path('/tmp/astrality/target1')
    template_target2 = Path('/tmp/astrality/target2')

    yield template_target1, template_target2

    if template_target1.is_file():
        os.remove(template_target1)
    if template_target2.is_file():
        os.remove(template_target2)

@pytest.mark.slow
def test_hot_reloading(
    test_template_targets,
    default_global_options,
    _runtime,
    test_config_directory
):
    template_target1, template_target2 = test_template_targets
    config1 = test_config_directory / 'astrality1.yml'
    config2 = test_config_directory / 'astrality2.yml'
    target_config = test_config_directory / 'astrality.yml'
    temp_directory = Path('/tmp/astrality')

    # Copy the first configuration into place
    shutil.copy(str(config1), str(target_config))

    application_config1 = dict_from_config_file(
        config1,
        context={},
    )
    application_config1.update(default_global_options)
    application_config1.update(_runtime)
    application_config1['config/astrality']['hot_reload_config'] = True

    module_manager = ModuleManager(application_config1)

    # Before beginning, the template should not be compiled
    assert not template_target1.is_file()

    # But when we finalize tasks, it should be compiled
    module_manager.finish_tasks()
    assert template_target1.is_file()

    # Also check that the filewatcher has been started
    assert module_manager.directory_watcher.observer.is_alive()

    # We now "edit" the configuration file
    shutil.copy(str(config2), str(target_config))
    time.sleep(0.7)

    # Since hot reloading is enabled, the new template target should be
    # compiled, and the old one cleaned up
    assert template_target2.is_file()
    assert not template_target1.is_file()

    # And we switch back again
    shutil.copy(str(config1), str(target_config))
    time.sleep(0.7)
    assert template_target1.is_file()
    assert not template_target2.is_file()

    # Cleanup config file
    if target_config.is_file():
        os.remove(target_config)

@pytest.yield_fixture
def three_watchable_files(test_config_directory):
    file1 = test_config_directory / 'file1.tmp'
    file2 = test_config_directory / 'file2.tmp'
    file3 = test_config_directory / 'file3.tmp'

    # Delete files if they are present
    if file1.is_file():
        os.remove(file1)
    if file2.is_file():
        os.remove(file2)
    if file3.is_file():
        os.remove(file3)

    yield file1, file2, file3

    # Delete files on cleanup
    if file1.is_file():
        os.remove(file1)
    if file2.is_file():
        os.remove(file2)
    if file3.is_file():
        os.remove(file3)

def test_all_three_actions_in_on_modified_block(
    three_watchable_files,
    default_global_options,
    _runtime,
    test_config_directory,
):
    file1, file2, file3 = three_watchable_files
    car_template = test_config_directory / 'templates' / 'a_car.template'
    mercedes_context = test_config_directory / 'context' / 'mercedes.yml'
    tesla_context = test_config_directory / 'context' / 'tesla.yml'

    application_config = {
        'module/car': {
            'on_startup': {
                'import_context': {
                    'from_path': str(mercedes_context),
                },
                'compile': {
                    'template': str(car_template),
                    'target': str(file1),
                },
            },
            'on_modified': {
                str(file2): {
                    'import_context': {
                        'from_path': str(tesla_context),
                    },
                    'compile': {
                        'template': str(car_template),
                        'target': str(file1),
                    },
                    'run': 'touch ' + str(file3),
                },
            },
        },
    }
    application_config.update(default_global_options)
    application_config.update(_runtime)

    module_manager = ModuleManager(application_config)

    # Sanity check before beginning testing
    assert not file1.is_file()
    assert not file2.is_file()
    assert not file3.is_file()

    # Now finish tasks, i.e. on_startup block
    module_manager.finish_tasks()
    assert file1.is_file()
    assert not file2.is_file()
    assert not file3.is_file()

    # Check that the correct context is inserted
    with open(file1) as file:
        assert file.read() == 'My car is a Mercedes'

    # Now modify file2 such that the on_modified block is triggered
    file2.write_text('some new content')
    time.sleep(0.7)

    # The on_modified run command should now have been executed
    assert file3.is_file()

    module_manager.exit()

def test_recompile_templates_when_modified(
    three_watchable_files,
    default_global_options,
    _runtime,
):
    template, target, _ = three_watchable_files
    template.touch()

    application_config = {
        'module/module_name': {
            'on_startup': {
                'compile': {
                    'template': str(template),
                    'target': str(target),
                },
            },
        },
        'context/section': {
            1: 'value',
        },
    }
    application_config.update(default_global_options)
    application_config.update(_runtime)
    application_config['config/modules'] = {
        'recompile_modified_templates': True,
    }

    module_manager = ModuleManager(application_config)

    # Sanity check before beginning testing
    with open(template) as file:
        assert file.read() == ''

    assert not target.is_file()

    # Compile the template
    module_manager.finish_tasks()
    with open(target) as file:
        assert file.read() == ''

    # Now write to the template and see if it is recompiled
    template.write_text('{{ section.2 }}')
    time.sleep(0.7)
    with open(target) as file:
        assert file.read() == 'value'

    module_manager.exit()


def test_recompile_templates_when_modified_overridden(
    three_watchable_files,
    default_global_options,
    _runtime,
):
    """
    If a file is watched in a on_modified block, it should override the
    recompile_modified_templates option.
    """
    template, target, touch_target = three_watchable_files
    template.touch()

    application_config = {
        'module/module_name': {
            'on_startup': {
                'compile': {
                    'template': str(template),
                    'target': str(target),
                },
            },
            'on_modified': {
                str(template): {
                    'run': 'touch ' + str(touch_target),
                },
            },
        },
        'context/section': {
            1: 'value',
        },
    }
    application_config.update(default_global_options)
    application_config.update(_runtime)
    application_config['config/modules'] = {
        'recompile_modified_templates': True,
    }

    module_manager = ModuleManager(application_config)

    # Sanity check before beginning testing
    with open(template) as file:
        assert file.read() == ''

    assert not target.is_file()

    # Compile the template
    module_manager.finish_tasks()
    with open(target) as file:
        assert file.read() == ''

    # Now write to the template and see if it is *compiled*, but the on_modified
    # command is run instead
    template.write_text('{{ section.2 }}')
    time.sleep(0.7)
    with open(target) as file:
        assert file.read() == ''
    assert touch_target.is_file()

    module_manager.exit()


def test_importing_context_on_modification(
    three_watchable_files,
    default_global_options,
    _runtime,
    test_config_directory,
):
    """Test that context values are imported in on_modified blocks."""
    file1, *_ = three_watchable_files
    mercedes_context = test_config_directory / 'context' / 'mercedes.yml'

    application_config = {
        'module/module_name': {
            'on_modified': {
                str(file1): {
                    'import_context': {
                        'from_path': str(mercedes_context),
                    },
                },
            },
        },
        'context/car': {
            'manufacturer': 'Tesla',
        },
    }
    application_config.update(default_global_options)
    application_config.update(_runtime)

    module_manager = ModuleManager(application_config)
    module_manager.finish_tasks()

    # Sanity check before modifying file1
    assert module_manager.application_context['car']['manufacturer'] == 'Tesla'

    # After modifying file1, Mercedes should have been imported
    file1.touch()
    file1.write_text('new content, resulting in importing Mercedes')
    time.sleep(0.7)
    assert module_manager.application_context['car']['manufacturer'] == 'Mercedes'



import shutil

import pytest

from astrality.module import GlobalModulesConfig, Module, ModuleManager

def test_that_module_configuration_is_processed_correctly_before_use(
    test_config_directory,
):
    """
    Test that all list item configurations can be given as single strings,
    and that missing configuration options are inserted.
    """
    module_config = {'module/A': {
        'on_startup': {
            'run': 'echo hi!',
        },
        'on_event': {
            'import_context': {'from_file': '/test'},
            'run': ['echo 1', 'echo 2'],
            'trigger': 'on_modified:/some/file',
        },
        'on_modified': {
            '/some/file': {
                'compile': {'template': '/some/template'},
            },
        },
    }}

    module = Module(
        module_config=module_config,
        module_directory=test_config_directory,
    )

    processed_config = {
        'on_startup': {
            'run': ['echo hi!'],
            'compile': [],
            'import_context': [],
            'trigger': [],
        },
        'on_event': {
            'import_context': [{'from_file': '/test'}],
            'run': ['echo 1', 'echo 2'],
            'compile': [{'template': '/some/template'}],
            'trigger': ['on_modified:/some/file'],
        },
        'on_exit': {
            'run': [],
            'compile': [],
            'import_context': [],
            'trigger': [],
        },
        'on_modified': {
            '/some/file': {
                'compile': [{'template': '/some/template'}],
                'run': [],
                'import_context': [],
                'trigger': [],
            },
        },
    }
    assert module.module_config == processed_config


def test_modules_config_explicitly_enabled_modules(
    test_config_directory,
):
    global_modules_config_dict = {
        'modules_directory': 'test_modules',
        'enabled_modules': [
            {'name': 'burma::burma'},
            {'name': 'india'},
        ],
    }
    global_modules_config = GlobalModulesConfig(
        config=global_modules_config_dict,
        config_directory=test_config_directory,
    )
    global_modules_config.compile_config_files({})

    assert 'burma::burma' in global_modules_config.enabled_modules
    assert 'india' in global_modules_config.enabled_modules
    assert 'whatever' not in global_modules_config.enabled_modules


def test_modules_config_implicitly_enabled_modules(
    test_config_directory,
):
    global_modules_config_dict = {
        'modules_directory': 'test_modules',
        'enabled_modules': [
            {'name': 'burma::*'},
            {'name': 'india'},
        ],
    }
    global_modules_config = GlobalModulesConfig(
        config=global_modules_config_dict,
        config_directory=test_config_directory,
    )
    global_modules_config.compile_config_files({})

    assert 'burma::burma' in global_modules_config.enabled_modules
    assert 'india' in global_modules_config.enabled_modules
    assert 'burma::only_defined_accepted' not in global_modules_config.enabled_modules

def test_modules_config_several_implicitly_enabled_modules(
    test_config_directory,
):
    global_modules_config_dict = {
        'modules_directory': 'test_modules',
        'enabled_modules': [
            {'name': 'two_modules::*'},
            {'name': 'india'},
        ],
    }
    global_modules_config = GlobalModulesConfig(
        config=global_modules_config_dict,
        config_directory=test_config_directory,
    )
    global_modules_config.compile_config_files({})

    assert 'two_modules::bhutan' in global_modules_config.enabled_modules
    assert 'two_modules::bangladesh' in global_modules_config.enabled_modules

def test_modules_config_where_all_modules_are_enabled(
    test_config_directory,
):
    global_modules_config_dict = {
        'modules_directory': str(test_config_directory / 'test_modules'),
        'enabled_modules': [
            {'name': '*::*'},
            {'name': '*'},
        ],
    }
    global_modules_config = GlobalModulesConfig(
        config=global_modules_config_dict,
        config_directory=test_config_directory,
    )
    global_modules_config.compile_config_files(
        {'module': {'setting': 'whatever'}},
    )

    assert 'two_modules::bhutan' in global_modules_config.enabled_modules
    assert 'two_modules::bangladesh' in global_modules_config.enabled_modules


def test_enabling_of_modules_defined_different_places(
    default_global_options,
    _runtime,
):
    application_config = {
        'config/modules': {
            'modules_directory': 'freezed_modules',
            'enabled_modules': [
                {'name': 'south_america::brazil'},
                {'name': 'india'},
            ],
        },
        'module/india': {},     # Enabled
        'module/pakistan': {},  # Not enabled
    }
    application_config.update(default_global_options)
    application_config.update(_runtime)

    module_manager = ModuleManager(application_config)

    assert len(module_manager.modules) == 2
    assert 'south_america::brazil' in module_manager.modules
    assert 'india' in module_manager.modules

def test_enabling_of_all_modules(
    default_global_options,
    _runtime,
):
    application_config = {
        'config/modules': {
            'modules_directory': 'freezed_modules',
        },
        'module/india': {},     # Enabled
        'module/pakistan': {},  # Not enabled
    }
    application_config.update(default_global_options)
    application_config.update(_runtime)

    module_manager = ModuleManager(application_config)

    assert len(module_manager.modules) == 5
    assert 'india' in module_manager.modules
    assert 'pakistan' in module_manager.modules
    assert 'north_america::USA' in module_manager.modules
    assert 'south_america::brazil' in module_manager.modules
    assert 'south_america::argentina' in module_manager.modules


@pytest.yield_fixture
def delete_jakobgm(test_config_directory):
    """Delete jakobgm module directory used in testing."""
    location = test_config_directory / 'freezed_modules' / 'jakobgm'

    yield

    if location.is_dir():
        shutil.rmtree(location)

def test_using_three_different_module_sources(
    default_global_options,
    _runtime,
    test_config_directory,
    delete_jakobgm,
):
    modules_directory = test_config_directory / 'freezed_modules'

    application_config = {
        'config/modules': {
            'modules_directory': str(modules_directory),
            'enabled_modules': [
                {'name': 'north_america::*'},
                {'name': 'github::jakobgm/test-module.astrality'},
                {'name': 'italy'},
            ],
        },
        'module/italy': {},
        'module/spain': {},
    }
    application_config.update(default_global_options)
    application_config.update(_runtime)

    module_manager = ModuleManager(application_config)

    assert len(module_manager.modules) == 4
    assert 'north_america::USA' in module_manager.modules
    assert 'github::jakobgm/test-module.astrality::botswana' in module_manager.modules
    assert 'github::jakobgm/test-module.astrality::ghana' in module_manager.modules
    assert 'italy' in module_manager.modules



from datetime import datetime
import os
import time

import pytest

from astrality.module import ModuleManager


@pytest.yield_fixture
def three_temporary_files(test_config_directory):
    file1 = test_config_directory / 'file1.tmp'
    file2 = test_config_directory / 'file2.tmp'
    file3 = test_config_directory / 'file3.tmp'

    yield file1, file2, file3

    if file1.is_file():
        os.remove(file1)
    if file2.is_file():
        os.remove(file2)
    if file3.is_file():
        os.remove(file3)


def test_that_only_changed_events_are_run(
    default_global_options,
    _runtime,
    three_temporary_files,
    freezer,
):
    file1, file2, file3 = three_temporary_files

    application_config = {
        'module/weekday': {
            'event_listener': {'type': 'weekday'},
            'on_event': {'run': 'touch ' + str(file1)},
        },
        'module/periodic': {
            'event_listener': {'type': 'periodic', 'days': 1, 'hours': 12},
            'on_event': {'run': 'touch ' + str(file2)},
        },
    }
    application_config.update(default_global_options)
    application_config.update(_runtime)

    # Move to a monday
    freezer.move_to('2018-02-19')

    module_manager = ModuleManager(application_config)
    module_manager.finish_tasks()

    # No on_event should have been run
    assert not file1.is_file()
    assert not file2.is_file()

    # Move to a next tuesday, on day ahead
    freezer.move_to('2018-02-20')
    module_manager.finish_tasks()

    # Now the weekday module should have been run, but *not* the periodic one
    assert file1.is_file()
    assert not file2.is_file()

    # Delete the created file
    os.remove(file1)

    # Move 13 hours ahead, now the periodic timer should have been run
    freezer.move_to(datetime(
        year=2018,
        month=2,
        day=20,
        hour=13,
    ))
    module_manager.finish_tasks()

    # The periodic file should now have been created, but not the weekday one
    assert not file1.is_file()
    assert file2.is_file()

    # Now delete the two files and check if both events are detected at the
    # same time
    os.remove(file2)
    freezer.move_to('2020-01-01')
    module_manager.finish_tasks()
    assert file1.is_file()
    assert file2.is_file()



"""General utility functions which are used across the application."""

import logging
import os
import subprocess
from pathlib import Path
from typing import Any, Dict, Union

logger = logging.getLogger('astrality')


def run_shell(
    command: str,
    timeout: Union[int, float] = 2,
    fallback: Any = '',
    working_directory: Path = Path.home(),
    allow_error_codes: bool = False,
) -> str:
    """
    Return the standard output of a shell command.

    If the shell command has a non-zero exit code or times out, the function
    returns the `fallback` argument instead of the standard output.
    """

    process = subprocess.Popen(
        command,
        cwd=working_directory,
        shell=True,
        universal_newlines=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    try:
        # We add just a small extra wait in case users specify 0 seconds,
        # in order to not print an error when a command is really quick.
        if timeout == 0:
            process.wait(timeout=timeout + 0.1)
        else:
            process.wait(timeout=timeout)

        for error_line in process.stderr:
            logger.error(str(error_line))

        if process.returncode != 0 and not allow_error_codes:
            logger.error(
                f'Command "{command}" exited with non-zero return code: ' +
                str(process.returncode)
            )
            return fallback
        else:
            stdout = process.communicate()[0]
            logger.info(stdout)
            return stdout.replace('\n', '')

    except subprocess.TimeoutExpired:
        logger.warning(
            f'The command "{command}" used more than {timeout} seconds in '
            'order to finish. The exit code can not be verified. This might be '
            'intentional for background processes and daemons.'
        )
        return fallback


def generate_expanded_env_dict() -> Dict[str, str]:
    """Return os.environ dict with all env variables expanded."""

    env_dict = {}
    for name, value in os.environ.items():
        try:
            env_dict[name] = os.path.expandvars(value)
        except ValueError as e:
            if 'invalid interpolation syntax' in str(e):
                logger.warning(f'''
                Could not use environment variable {name}={value}.
                It is too complex for expansion, using unexpanded value
                instead...
                ''')
                env_dict[name] = value
            else:
                raise

    return env_dict





import logging
import os
from pathlib import Path
from typing import Union

from astrality.exceptions import GithubModuleError
from astrality.utils import run_shell

def clone_repo(
    user: str,
    repository: str,
    modules_directory: Path,
    timeout: Union[int, float] = 50,
) -> Path:
    """
    Clone Github `user`/`repository` to modules_directory.

    The resulting repository is placed in:
    <modules_directory>/<user>/<repository>.
    """

    github_user_directory = modules_directory / user
    github_user_directory.mkdir(parents=True, exist_ok=True)
    repository_directory = github_user_directory / repository
    github_url = f'https://github.com/{user}/{repository}.git {repository_directory}'

    # Fail on git credential prompt: https://serverfault.com/a/665959
    result = run_shell(
        command='GIT_TERMINAL_PROMPT=0 git clone ' + github_url,
        timeout=timeout,
        fallback=False,
        working_directory=github_user_directory,
        allow_error_codes=True,
    )

    if not repository_directory.is_dir() or result is False:
        try:
            github_user_directory.rmdir()
        except OSError:
            pass

        raise GithubModuleError(
            f'Could not clone repository "{user}/{repository}.\n'
            f'Return value from cloning operation: "{result}".'
        )

    return repository_directory


def clone_or_pull_repo(
    user: str,
    repository: str,
    modules_directory: Path,
    timeout: Union[int, float] = 50,
) -> Path:
    github_repo_directory = modules_directory / user / repository

    if not github_repo_directory.is_dir():
        # The repository does not exist, so we clone it
        return clone_repo(
            user=user,
            repository=repository,
            modules_directory=modules_directory,
            timeout=timeout,
        )

    logger = logging.getLogger(__name__)
    if not (github_repo_directory / '.git').is_dir():
        logger.error(
            f'Tried to update git module directory "{github_repo_directory}", '
            'but the directory does not contain a ".git" sub-directory.'
        )
        return github_repo_directory

    result = run_shell(
        command='GIT_TERMINAL_PROMPT=0 git pull',
        timeout=timeout,
        fallback=False,
        working_directory=github_repo_directory,
        allow_error_codes=True,
    )
    if result is False:
        raise GithubModuleError(
            f'Could not git pull module directory "{github_repo_directory}".\.'
            f'Return value from git pull operation: "{result}".'
        )

    return github_repo_directory



"""Test module for the use of external modules."""
import os
import time

import pytest

from astrality.module import ModuleManager, GlobalModulesConfig


def test_that_external_modules_are_brought_in(
    test_config_directory,
    default_global_options,
    _runtime,
):
    application_config = {
        'config/modules': {
            'modules_directory': 'test_modules',
            'enabled_modules': [
                {'name': 'thailand::thailand'},
                {'name': '*'},
            ],
        },
        'module/cambodia': {
            'enabled_modules': True,
        },
    }
    application_config.update(default_global_options)
    application_config.update(_runtime)

    module_manager = ModuleManager(application_config)

    thailand_path = test_config_directory / 'test_modules' / 'thailand'
    assert tuple(module_manager.modules.keys()) == (
        'thailand::thailand',
        'cambodia',
    )


@pytest.yield_fixture
def temp_test_files(test_config_directory):
    module_dir = test_config_directory / 'test_modules' / 'using_all_actions'
    watched_file = module_dir / 'watched_for_modifications'

    compile_target = module_dir / 'compiled.tmp'
    touch_target = module_dir / 'touched.tmp'
    watch_touch_target = module_dir / 'watch_touched.tmp'

    for file in (compile_target, touch_target, watch_touch_target,):
        if file.is_file():
            os.remove(file)

    yield compile_target, touch_target, watch_touch_target, watched_file

    for file in (compile_target, touch_target, watch_touch_target,):
        if file.is_file():
            os.remove(file)


def test_correct_relative_paths_used_in_external_module(
    temp_test_files,
    test_config_directory,
    default_global_options,
    _runtime,
):
    application_config = {
        'config/modules': {
            'modules_directory': 'test_modules',
            'enabled_modules': [{'name': 'using_all_actions::*'}],
        },
    }
    application_config.update(default_global_options)
    application_config.update(_runtime)

    module_manager = ModuleManager(application_config)

    compile_target, touch_target, watch_touch_target, watched_file = temp_test_files

    # Sanity check before testing
    for file in (compile_target, touch_target, watch_touch_target,):
        assert not file.is_file()

    # Finish task and see if context import, compilation, and run has been
    # correctly run relative to the module directory path
    module_manager.finish_tasks()
    with open(compile_target, 'r') as file:
        assert file.read() == "Vietnam's capitol is Ho Chi Minh City"
    assert touch_target.is_file()

    # Now modify the observed file, and see if on_modified is triggered
    watched_file.write_text('This watched file has been modified')
    time.sleep(0.7)

    with open(compile_target, 'r') as file:
        assert file.read() == "Vietnam's capitol is Hanoi"
    assert watch_touch_target.is_file()


def test_that_external_module_contexts_are_imported_correctly(
    test_config_directory,
    default_global_options,
    _runtime,
):
    application_config = {
        'config/modules': {
            'modules_directory': 'test_modules',
            'enabled_modules': [{'name': 'module_with_context::*', }],
        },
        'context/china': {
            'capitol': 'beijing',
        },
    }
    application_config.update(default_global_options)
    application_config.update(_runtime)

    module_manager = ModuleManager(application_config)

    expected_context = {
        'laos': {'capitol': 'vientiane'},
        'china': {'capitol': 'beijing'},
    }
    assert len(module_manager.application_context) == 2
    for key, value in expected_context.items():
        assert module_manager.application_context[key] == value
